# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import randint

def get_wait_time(retry_num):
    """
    Compute time for exponential backoff.

    Args:
        retry_num (int): Retry attempt number to determine wait time.

    Returns:
        (int): Amount of time to wait.
    """
    return 2 ** (retry_num+3)

def get_wait_time_rand(retry_num):
    """
    Compute time for exponential backoff using a random element.

    Args:
        retry_num (int): Retry attempt number to determine random number range.

    Returns:
        (int): Amount of time to wait.
    """
    return randint(1, 2 ** (retry_num+3))



#!/usr/bin/env python
# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ingestclient.core.engine import Engine
from ingestclient.core.config import ConfigFileError
from ingestclient.core.backend import BossBackend
from ingestclient import check_version
from ingestclient.utils.log import always_log_info
from ingestclient.utils.console import print_estimated_job

from six.moves import input
import datetime
import argparse
import sys
import multiprocessing as mp
import os
import time
import logging


def get_confirmation(prompt, force=False):
    """Method to confirm decisions

    Args:
        prompt(str): Question to ask the user
        force(bool): Flag indicating if user prompts should be ignored

    Returns:
        (bool): True indicating yes, False indicating no
    """
    if not force:
        decision = False
        while True:
            confirm = input("{} (y/n): ".format(prompt))
            if confirm.lower() == "y":
                decision = True
                break
            elif confirm.lower() == "n":
                decision = False
                break
            else:
                print("Enter 'y' or 'n' for 'yes' or 'no'")

        return decision
    else:
        return True


def worker_process_run(api_token, job_id, pipe, config_file=None, configuration=None):
    """A worker process main execution function. Generates an engine, and joins the job
       (that was either created by the main process or joined by it).
       Ends when no more tasks are left that can be executed.

    Args:
        api_token(str): the token to initialize the engine with.
        job_id(int): the id of the job the engine needs to join with.
        pipe(multiprocessing.Pipe): the receiving end of the pipe that communicates with the master process.
        config_file(str): the path to the configuration file (configuration required if omitted)
        configuration(Configuration): a pre-loaded configuration object (config_file required if omitted)

    """
    always_log_info("Creating new worker process, pid={}.".format(os.getpid()))

    # Create the engine
    if config_file is None and configuration is None:
        raise Exception('Must provide either a configuration instance or a configuration file')

    try:
        engine = Engine(config_file=config_file, 
                        configuration=configuration,
                        backend_api_token=api_token, 
                        ingest_job_id=job_id)
    except ConfigFileError as err:
        print("ERROR (pid: {}): {}".format(os.getpid(), err))
        sys.exit(1)

    # Join job
    engine.join()

    # Start it up!
    should_run = True
    while should_run:
        try:
            engine.run()
            # run will end if no more jobs are available
            should_run = False
        except KeyboardInterrupt:
            # Make sure they want to stop this client, wait for the main process to send the next step
            should_run = pipe.recv()
    always_log_info("  - Process pid={} finished gracefully.".format(os.getpid()))
    

def get_parser():
    parser = argparse.ArgumentParser(description="Client for facilitating large-scale data ingest",
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog="Visit https://docs.theBoss.io for more details")

    parser.add_argument("--api-token", "-a",
                        default=None,
                        help="Token for API authentication. If not provided and ndio is configured those credentials will automatically be used.")
    parser.add_argument("--job-id", "-j",
                        default=None,
                        help="ID of the ingest job if joining an existing ingest job")
    parser.add_argument("--log-file", "-l",
                        default=None,
                        help="Absolute path to the logfile to use")
    parser.add_argument("--log-level", "-v",
                        default="warning",
                        help="Log level to use: critical, error, warning, info, debug")
    parser.add_argument("--version",
                        action="store_true",
                        default=False,
                        help="Get the package version")
    parser.add_argument("--cancel", "-c",
                        action="store_true",
                        default=None,
                        help="Flag indicating if you'd like to cancel (and remove) an ingest job. This will not delete data already ingested, but will prevent continuing this ingest job.")
    parser.add_argument("--force", "-f",
                        action="store_true",
                        default=False,
                        help="Flag indicating if you'd like to ignore all confirmation prompts.")
    parser.add_argument("--manual-complete", "-m",
                        action="store_true",
                        default=False,
                        help="Flag indicating if you want to manually mark an Ingest Job for completion. If omitted, the client will automatically cleanup after a successful upload")
    parser.add_argument("--processes_nb", "-p", type=int,
                        default=1,
                        help="The number of client processes that will upload the images of the ingest job.")
    parser.add_argument("config_file", nargs='?', help="Path to the ingest job configuration file")

    return parser


def start_workers(ingest_job_id, args, configuration):
    """
    Start upload processes.

    Args:
        ingest_job_id (int):
        args (Namespace): Command line arguments.
        configuration(ingestclient.core.config.Configuration): A pre-loaded configuration instance.

    Returns:
        (list[(Process, Pipe)]): list of processes and their pipes.
    """
    # Ceate worker processes
    workers = []
    for i in range(args.processes_nb):
        new_pipe = mp.Pipe(False)
        new_process = mp.Process(target=worker_process_run, 
                                 args=(args.api_token, ingest_job_id, new_pipe[0]),
                                 kwargs={'config_file': args.config_file, 'configuration': configuration}
                                 )
        workers.append((new_process, new_pipe[1]))
        new_process.start()

        # Sleep to slowly ramp up load on lambda
        time.sleep(.5)

    return workers


def upload(engine, args, configuration, start_time):
    """
    Kick off upload processes and monitor them.

    Args:
        engine (ingestclient.core.Engine):
        args (Namespace): Command line arguments.
        configuration(ingestclient.core.config.Configuration): A pre-loaded configuration instance.
        start_time (float): When ingest client started in seconds since 1/1/1970 00:00.

    Returns:
        (bool): True if uploading not finished.
    """
    workers = start_workers(engine.ingest_job_id, args, configuration)

    # Start the main process engine
    should_run = True
    job_complete = False
    while should_run:
        try:
            engine.monitor(workers)
            # run will end if no more jobs are available, join other processes
            should_run = False
            status = engine.backend.get_job_status(engine.ingest_job_id)
            if status:
                if status["current_message_count"] == 0:
                    job_complete = True
                else:
                    always_log_info("Error: Something has gone wrong, monitor has ended even though there are still messages in upload queue.")
            else:
                always_log_info("Unable to get job status - not marking job as complete.")

        except KeyboardInterrupt:
            # Make sure they want to stop this client
            while True:
                quit_uploading = input("Are you sure you want to quit uploading? (y/n)")
                if quit_uploading.lower() == "y":
                    always_log_info("Stopping upload engine.")
                    should_run = False
                    break
                elif quit_uploading.lower() == "n":
                    print("Continuing...")
                    break
                else:
                    print("Enter 'y' or 'n' for 'yes' or 'no'")

            # notify the worker processes that they should stop execution
            for _, worker_pipe in workers:
                worker_pipe.send(should_run)

    always_log_info("Waiting for worker processes to close...\n")
    time.sleep(1)  # Make sure workers have cleaned up
    for worker_process, worker_pipe in workers:
        worker_process.join()
        worker_pipe.close()

    if job_complete:
        # If auto-complete, mark the job as complete and cleanup
        always_log_info("All upload tasks completed in {:.2f} minutes.".format((time.time() - start_time) / 60))
        if not args.manual_complete:
            always_log_info(" - Marking Ingest Job as complete and cleaning up. Please wait.")
            if not engine.complete():
                always_log_info("Unable to complete, still have chunks that aren't ingested")
                return True
            always_log_info(" - Cleanup Done")
        else:
            always_log_info(" - Auto-complete disabled. This ingest job will remain in the 'Uploading' state until you manually mark it as complete")
            return False
    else:
        always_log_info("Client exiting")
        always_log_info("Run time: {:.2f} minutes.".format((time.time() - start_time) / 60))
        return False

def main(configuration=None, parser_args=None):
    """Client UI main

    Args:
        configuration(ingestclient.core.config.Configuration): A pre-loaded configuration instance
        parser_args(argparse.ArgumentParser): A pre-loaded ArgumentParser instance

    Returns:

    """
    parser = get_parser()
    if parser_args is None:
        args = parser.parse_args()
    else:
        args = parser_args

    # Get the version
    if args.version:
        check_version()
        return

    # Make sure you have a config file
    if args.config_file is None and configuration is None:
        if args.cancel:
            # If no config is provided and you are deleting, the client defaults to the production Boss stack
            boss_backend_params = {"client": {
                "backend": {
                    "name": "boss",
                    "class": "BossBackend",
                    "host": "api.theboss.io",
                    "protocol": "https"}}}
            backend = BossBackend(boss_backend_params)
            backend.setup(args.api_token)

            # Trying to cancel
            if args.job_id is None:
                parser.print_usage()
                print("Error: You must provide an ingest job ID to cancel")
                sys.exit(1)

            if not get_confirmation("Are you sure you want to cancel ingest job {}? ".format(args.job_id), args.force):
                print("Command ignored. Job not cancelled")
                sys.exit(0)

            backend.cancel(args.job_id)
            print("Ingest job {} successfully cancelled.".format(args.job_id))
            sys.exit(0)
        else:
            # Not deleting, so you need a config file
            parser.print_usage()
            print("Error: Ingest Job Configuration File is required")
            sys.exit(1)

    # Setup logging
    log_level = logging.getLevelName(args.log_level.upper())
    if not args.log_file:
        # Using default log path
        log_path = os.path.expanduser("~/.boss-ingest")
        log_file = os.path.join(log_path,
                                'ingest_log{}_pid{}.log'.format(datetime.datetime.now().strftime("%Y%m%d-%H%M%S"),
                                                                os.getpid()))
        # Make sure the logs dir exists if using the default log path
        if not os.path.exists(log_path):
            os.makedirs(log_path)
    else:
        log_file = args.log_file

    logging.basicConfig(level=log_level,
                        format='%(asctime)s %(levelname)-8s %(message)s',
                        datefmt='%m-%d %H:%M',
                        filename=log_file,
                        filemode='a')
    logging.getLogger('ingest-client').addHandler(logging.StreamHandler(sys.stdout))

    # Create an engine instance
    try:
        engine = Engine(config_file=args.config_file, 
                        backend_api_token=args.api_token, 
                        ingest_job_id=args.job_id, 
                        configuration=configuration)
    except ConfigFileError as err:
        print("ERROR: {}".format(err))
        sys.exit(1)

    if args.cancel:
        # Trying to cancel
        if args.job_id is None:
            parser.print_usage()
            print("Error: You must provide an ingest job ID to cancel")
            sys.exit(1)

        if not get_confirmation("Are you sure you want to cancel ingest job {}? ".format(args.job_id), args.force):
            print("Command ignored. Job not cancelled")
            sys.exit(0)

        always_log_info("Attempting to cancel Ingest Job {}.".format(args.job_id))
        engine.cancel()
        always_log_info("Ingest job {} successfully cancelled.".format(args.job_id))
        sys.exit(0)

    else:
        # Trying to create or join an ingest
        if args.job_id is None:
            # Creating a new session - make sure the user wants to do this.
            print_estimated_job(config_file=args.config_file, configuration=configuration)
            print("\n")
            if not get_confirmation("Would you like to create a NEW ingest job?", args.force):
                # Don't want to create a new job
                print("Exiting")
                sys.exit(0)
        else:
            # Resuming a session - make sure the user wants to do this.
            if not get_confirmation("Are you sure you want to resume ingest job {}?".format(args.job_id), args.force):
                # Don't want to resume
                print("Exiting")
                sys.exit(0)

    # Setup engine instance.  Prompt user to confirm things if needed
    question_msgs = engine.setup()
    if question_msgs:
        for msg in question_msgs:
            if not get_confirmation(msg, args.force):
                print("Ingest job cancelled")
                sys.exit(0)

    if args.job_id is None:
        # Create job
        engine.create_job()
        always_log_info("Successfully Created Ingest Job ID: {}".format(engine.ingest_job_id))
        always_log_info("Note: You need this ID to continue this job later!")

        if not get_confirmation("\nDo you want to start uploading now?", args.force):
            print("OK - Your job is waiting for you. You can resume by providing Ingest Job ID '{}' to the client".format(engine.ingest_job_id))
            sys.exit(0)

    # Join job
    engine.join()

    start_time = time.time()
    while upload(engine, args, configuration, start_time):
        pass



if __name__ == '__main__':
    main()



# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
import six
from PIL import Image
import re
import os
import h5py
import numpy as np
from math import floor
import botocore
import logging


from ..utils.filesystem import DynamicFilesystemAbsPath
from .path import PathProcessor
from .tile import TileProcessor


class Hdf5TimeSeriesPathProcessor(PathProcessor):
    """A Path processor for time-series, multi-channel data (e.g. calcium imaging)

    Assumes the data is stored (t,y,z, channel) in individual hdf5 files, with 1 hdf5 file per z-slice

    """
    def __init__(self):
        """Constructor to add custom class var"""
        PathProcessor.__init__(self)
        self.regex = None

    def setup(self, parameters):
        """Set the params

        MUST HAVE THE CUSTOM PARAMETERS: "root_dir": "<path_to_stack_root>",
                                         "extension": "hdf5|h5",
                                         "base_filename": the base filename, see below for how this is parsed,

        base_filename string identifies how to insert the z-index value into the filename. Identify a place to insert
        the z_index with "<>".  If you want to offset add o:number. If you want to zero pad add z:number"

        my_base_<> -> my_base_0, my_base_1, my_base_2
        <o:200>_my_base_<p:4> -> 200_my_base_0000, 201_my_base_0001, 202_my_base_0002

        Includes the "ingest_job" section of the config file automatically

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        self.parameters = parameters
        self.regex = re.compile('<(o:\d+)?(p:\d+)?>')

    def process(self, x_index, y_index, z_index, t_index=None):
        """
        Method to compute the file path for the indicated Z-slice

        Args:
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (str): An absolute file path that contains the specified data

        """
        if z_index >= self.parameters["ingest_job"]["extent"]["z"][1]:
            raise IndexError("Z-index out of range")

        # Create base filename
        matches = self.regex.findall(self.parameters['base_filename'])

        base_str = self.parameters['base_filename']
        for m in matches:
            if m[0]:
                # there is an offset
                z_val = int(m[0].split(':')[1]) + z_index
            else:
                z_val = z_index

            if m[1]:
                # There is zero padding
                z_str = str(z_val).zfill(int(m[1].split(':')[1]))
            else:
                z_str = str(z_val)

            base_str = base_str.replace("<{}{}>".format(m[0], m[1]), z_str)

        # prepend root, append extension
        return os.path.join(self.parameters['root_dir'], "{}.{}".format(base_str, self.parameters['extension']))


class Hdf5TimeSeriesTileProcessor(TileProcessor):
    """A Tile processor for time-series, multi-channel data (e.g. calcium imaging)

    Assumes the data is stored (t, x, y, channel) in individual hdf5 files, with 1 hdf5 file per z-slice
    where x is the column dim and y is the row dim

    """

    def __init__(self):
        """Constructor to add custom class var"""
        TileProcessor.__init__(self)
        self.fs = None

    def setup(self, parameters):
        """ Method to load the file for uploading

        Args:
            parameters (dict): Parameters for the dataset to be processed


        MUST HAVE THE CUSTOM PARAMETERS: "upload_format": "<png|tif>",
                                         "channel_index": integer,
                                         "scale_factor": float,
                                         "dataset": str,
                                         "filesystem": "<s3|local>",
                                         "bucket": (if s3 filesystem)

        Returns:
            None
        """
        self.parameters = parameters
        self.fs = DynamicFilesystemAbsPath(parameters['filesystem'], parameters)

    def process(self, file_path, x_index, y_index, z_index, t_index=0):
        """
        Method to load the image file. Can break the image into smaller tiles to help make ingest go smoother, but
        currently must be perfectly divisible

        Args:
            file_path(str): An absolute file path for the specified tile
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (io.BufferedReader): A file handle for the specified tile

        """
        file_path = self.fs.get_file(file_path)

        x_range = [self.parameters["ingest_job"]["tile_size"]["x"] * x_index,
                   self.parameters["ingest_job"]["tile_size"]["x"] * (x_index + 1)]
        y_range = [self.parameters["ingest_job"]["tile_size"]["y"] * y_index,
                   self.parameters["ingest_job"]["tile_size"]["y"] * (y_index + 1)]

        # Open hdf5
        h5_file = h5py.File(file_path, 'r')

        # Save sub-img to png and return handle
        tile_data = np.array(h5_file[self.parameters['dataset']][t_index,
                                                                 x_range[0]:x_range[1],
                                                                 y_range[0]:y_range[1],
                                                                 int(self.parameters['channel_index'])])

        tile_data = np.swapaxes(tile_data, 0, 1)
        tile_data = np.multiply(tile_data, self.parameters['scale_factor'])
        tile_data = tile_data.astype(np.uint16)
        upload_img = Image.fromarray(tile_data, 'I;16')

        output = six.BytesIO()
        upload_img.save(output, format=self.parameters["upload_format"].upper())

        # Send handle back
        return output


class Hdf5TimeSeriesLabelTileProcessor(TileProcessor):
    """A Tile processor for label data packed in a time-series, multi-channel HDF5 (e.g. ROIs for calcium imaging)

    Assumes the data is stored (x, y) in individual hdf5 files, with 1 hdf5 file per z-slice
        where x is the column dim and y is the row dim

    """

    def __init__(self):
        """Constructor to add custom class var"""
        TileProcessor.__init__(self)
        self.fs = None

    def setup(self, parameters):
        """ Method to load the file for uploading

        Args:
            parameters (dict): Parameters for the dataset to be processed


        MUST HAVE THE CUSTOM PARAMETERS: "upload_format": "<png|tif>",
                                         "dataset": str,
                                         "filesystem": "<s3|local>",
                                         "bucket": (if s3 filesystem)

        Returns:
            None
        """
        self.parameters = parameters
        self.fs = DynamicFilesystemAbsPath(parameters['filesystem'], parameters)

    def process(self, file_path, x_index, y_index, z_index, t_index=0):
        """
        Method to load the image file. Can break the image into smaller tiles to help make ingest go smoother, but
        currently must be perfectly divisible

        Args:
            file_path(str): An absolute file path for the specified tile
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (io.BufferedReader): A file handle for the specified tile

        """
        file_path = self.fs.get_file(file_path)

        x_range = [self.parameters["ingest_job"]["tile_size"]["x"] * x_index,
                   self.parameters["ingest_job"]["tile_size"]["x"] * (x_index + 1)]
        y_range = [self.parameters["ingest_job"]["tile_size"]["y"] * y_index,
                   self.parameters["ingest_job"]["tile_size"]["y"] * (y_index + 1)]

        # Open hdf5
        h5_file = h5py.File(file_path, 'r')

        # Save sub-img to png and return handle
        tile_data = np.array(h5_file[self.parameters['dataset']][x_range[0]:x_range[1], y_range[0]:y_range[1]])
        tile_data = np.swapaxes(tile_data, 0, 1)
        tile_data = tile_data.astype(np.uint32)
        upload_img = Image.fromarray(tile_data, 'I')

        output = six.BytesIO()
        upload_img.save(output, format=self.parameters["upload_format"].upper())

        # Send handle back
        return output


class Hdf5SlicePathProcessor(PathProcessor):
    """A Path processor for large slices stored in hdf5 files.

    Assumes the data is stored in a dataset and an optional offset is stored in a dataset

    """
    def __init__(self):
        """Constructor to add custom class var"""
        PathProcessor.__init__(self)
        self.regex = None

    def setup(self, parameters):
        """Set the params

        MUST HAVE THE CUSTOM PARAMETERS: "root_dir": "<path_to_stack_root>",
                                         "extension": "hdf5|h5",
                                         "base_filename": the base filename, see below for how this is parsed,

        base_filename string identifies how to insert the z-index value into the filename. Identify a place to insert
        the z_index with "<>".  If you want to offset add o:number. If you want to zero pad add z:number"

        my_base_<> -> my_base_0, my_base_1, my_base_2
        <o:200>_my_base_<p:4> -> 200_my_base_0000, 201_my_base_0001, 202_my_base_0002

        Includes the "ingest_job" section of the config file automatically

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        self.parameters = parameters
        self.regex = re.compile('<(o:\d+)?(p:\d+)?>')

    def process(self, x_index, y_index, z_index, t_index=None):
        """
        Method to compute the file path for the indicated Z-slice

        Args:
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (str): An absolute file path that contains the specified data

        """
        if z_index >= self.parameters["ingest_job"]["extent"]["z"][1]:
            raise IndexError("Z-index out of range")

        # Create base filename
        matches = self.regex.findall(self.parameters['base_filename'])

        base_str = self.parameters['base_filename']
        for m in matches:
            if m[0]:
                # there is an offset
                z_val = int(m[0].split(':')[1]) + z_index
            else:
                z_val = z_index

            if m[1]:
                # There is zero padding
                z_str = str(z_val).zfill(int(m[1].split(':')[1]))
            else:
                z_str = str(z_val)

            base_str = base_str.replace("<{}{}>".format(m[0], m[1]), z_str)

        # prepend root, append extension
        return os.path.join(self.parameters['root_dir'], "{}.{}".format(base_str, self.parameters['extension']))


class Hdf5SliceTileProcessor(TileProcessor):
    """A Tile processor for large slices stored in hdf5 files.

    Assumes the data is stored in a dataset and an optional offset is stored in a dataset

    """

    def __init__(self):
        """Constructor to add custom class var"""
        TileProcessor.__init__(self)
        self.fs = None

    def setup(self, parameters):
        """ Method to load the file for uploading

        Args:
            parameters (dict): Parameters for the dataset to be processed


        MUST HAVE THE CUSTOM PARAMETERS: "upload_format": "<png|tif>",
                                         "data_name": str,
                                         "offset_name": str,
                                         "extent_name": str,
                                         "offset_origin_x": int,
                                         "offset_origin_y": int,
                                         "filesystem": "<s3|local>",
                                         "bucket": (if s3 filesystem)

        Returns:
            None
        """
        self.parameters = parameters
        self.fs = DynamicFilesystemAbsPath(parameters['filesystem'], parameters)

    def process(self, file_path, x_index, y_index, z_index, t_index=0):
        """
        Method to load the image file.

        Args:
            file_path(str): An absolute file path for the specified tile
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (io.BufferedReader): A file handle for the specified tile

        """
        file_path = self.fs.get_file(file_path)

        # Compute global range
        tile_x_range = [self.parameters["ingest_job"]["tile_size"]["x"] * x_index,
                        self.parameters["ingest_job"]["tile_size"]["x"] * (x_index + 1)]
        tile_y_range = [self.parameters["ingest_job"]["tile_size"]["y"] * y_index,
                        self.parameters["ingest_job"]["tile_size"]["y"] * (y_index + 1)]

        # Open hdf5
        h5_file = h5py.File(file_path, 'r')

        # Compute range in actual data, taking offsets into account
        x_offset = h5_file[self.parameters['offset_name']][1]
        y_offset = h5_file[self.parameters['offset_name']][0]

        x_img_extent = h5_file[self.parameters['extent_name']][1]
        y_img_extent = h5_file[self.parameters['extent_name']][0]

        x_frame_offset = x_offset + self.parameters['offset_origin_x']
        y_frame_offset = y_offset + self.parameters['offset_origin_x']

        x1 = max(tile_x_range[0], x_frame_offset)
        y1 = max(tile_y_range[0], y_frame_offset)
        x2 = min(tile_x_range[1], x_frame_offset + x_img_extent)
        y2 = min(tile_y_range[1], y_frame_offset + y_img_extent)

        if self.parameters['datatype'] == "uint8":
            datatype = np.uint8
        elif self.parameters['datatype']== "uint16":
            datatype = np.uint16
        else:
            raise Exception("Unsupported datatype: {}".format(self.parameters['datatype']))

        # Allocate Tile
        tile_data = np.zeros((self.parameters["ingest_job"]["tile_size"]["y"],
                             self.parameters["ingest_job"]["tile_size"]["x"]),
                             dtype=datatype, order='C')

        # Copy sub-img to tile, save, return
        img_y_index_start = max(0, y1 - y_frame_offset)
        img_y_index_stop = max(0, y2 - y_frame_offset)

        img_x_index_start = max(0, x1 - x_frame_offset)
        img_x_index_stop = max(0, x2 - x_frame_offset)

        tile_data[y1-tile_y_range[0]:y2-tile_y_range[0],
                  x1 - tile_x_range[0]:x2 - tile_x_range[0]] = np.array(h5_file[self.parameters['data_name']][
                                                                        img_y_index_start:img_y_index_stop,
                                                                        img_x_index_start:img_x_index_stop])

        tile_data = tile_data.astype(datatype)
        upload_img = Image.fromarray(tile_data)

        output = six.BytesIO()
        upload_img.save(output, format=self.parameters["upload_format"].upper())

        # Send handle back
        return output


class Hdf5ChunkPathProcessor(PathProcessor):
    """A Path processor for chunks stored in hdf5 files.

    Assumes the data is stored in a dataset and the filename contains the loction. supports an xyz offset

    """
    def __init__(self):
        """Constructor to add custom class var"""
        PathProcessor.__init__(self)
        self.regex = None

    def setup(self, parameters):
        """Set the params

        MUST HAVE THE CUSTOM PARAMETERS: "root_dir": "<path_to_stack_root>",
                                         "extension": "hdf5|h5",
                                         "prefix": Prefix for the filename
                                         "x_offset": the offset from 0 in the x dim
                                         "y_offset": the offset from 0 in the y dim
                                         "z_offset": the offset from 0 in the z dim
                                         "x_chunk_size": the chunk extent in the x dimension
                                         "y_chunk_size": the chunk extent in the y dimension
                                         "z_chunk_size": the chunk extent in the z dimension
                                         "use_python_convention" <bool>: a flag indicating if ranges use python convention

        filename format: prefix_xstart-xstop_ystart-ystop_zstart-zstop.h5

        Includes the "ingest_job" section of the config file automatically

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        self.parameters = parameters

    def process(self, x_index, y_index, z_index, t_index=None):
        """
        Method to compute the file path for the indicated Z-slice

        Args:
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (str): An absolute file path that contains the specified data

        """
        xstart = (x_index * self.parameters['x_chunk_size']) + self.parameters['x_offset']
        xstop = ((x_index + 1) * self.parameters['x_chunk_size']) + self.parameters['x_offset']
        if not self.parameters['use_python_convention']:
            xstop -= 1

        ystart = (y_index * self.parameters['y_chunk_size']) + self.parameters['y_offset']
        ystop = ((y_index + 1) * self.parameters['y_chunk_size']) + self.parameters['y_offset']
        if not self.parameters['use_python_convention']:
            ystop -= 1

        zstart = floor((z_index + self.parameters['z_offset']) / self.parameters['z_chunk_size'])
        zstop = ((zstart + 1) * self.parameters['z_chunk_size']) + self.parameters['z_offset']
        if not self.parameters['use_python_convention']:
            zstop -= 1
        zstart += self.parameters['z_offset']

        # prepend root, append extension
        filename = "{}_{}-{}_{}-{}_{}-{}.{}".format(self.parameters['prefix'],
                                                    xstart, xstop,
                                                    ystart, ystop,
                                                    zstart, zstop,
                                                    self.parameters['extension'])
        return os.path.join(self.parameters['root_dir'], filename)


class Hdf5ChunkTileProcessor(TileProcessor):
    """A Tile processor for large slices stored in hdf5 files.

    Assumes the data is stored in a dataset and an optional offset is stored in a dataset

    """

    def __init__(self):
        """Constructor to add custom class var"""
        TileProcessor.__init__(self)
        self.fs = None

    def setup(self, parameters):
        """ Method to load the file for uploading

        Args:
            parameters (dict): Parameters for the dataset to be processed


        MUST HAVE THE CUSTOM PARAMETERS: "upload_format": "<png|tiff>",
                                         "data_name": str,
                                         "z_chunk_size": the chunk extent in the z dimension,
                                         "filesystem": "<s3|local>",
                                         "bucket": (if s3 filesystem)

        Returns:
            None
        """
        self.parameters = parameters
        self.fs = DynamicFilesystemAbsPath(parameters['filesystem'], parameters)

    def process(self, file_path, x_index, y_index, z_index, t_index=0):
        """
        Method to load the image file.

        Args:
            file_path(str): An absolute file path for the specified tile
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (io.BufferedReader): A file handle for the specified tile

        """
        if self.parameters['datatype'] == "uint8":
            datatype = np.uint8
        elif self.parameters['datatype'] == "uint16":
            datatype = np.uint16
        elif self.parameters['datatype'] == "uint32":
            datatype = np.uint32
        else:
            raise Exception("Unsupported datatype: {}".format(self.parameters['datatype']))

        try:
            file_path = self.fs.get_file(file_path)

            # Open hdf5
            h5_file = h5py.File(file_path, 'r')

            # Compute z-index (plugin assumes xy extent fits in a tile)
            z_index = z_index % self.parameters['z_chunk_size']

            # Allocate Tile
            tile_data = np.array(h5_file[self.parameters['data_name']][z_index, :, :], dtype=datatype, order='C')

        except botocore.exceptions.ClientError as err:
            logger = logging.getLogger('ingest-client')
            logger.info("Could not find chunk. Assuming it's missing and generating blank data.")
            # TODO: remove kludge once we have contiguous datasets.
            tile_data = np.zeros((512, 512), dtype=datatype, order="C")
        except OSError as err:
            logger = logging.getLogger('ingest-client')
            logger.info("Could not find chunk. Assuming it's missing and generating blank data.")
            # TODO: remove kludge once we have contiguous datasets.
            tile_data = np.zeros((512, 512), dtype=datatype, order="C")

        upload_img = Image.fromarray(tile_data)
        output = six.BytesIO()
        upload_img.save(output, format=self.parameters["upload_format"].upper())

        # Send handle back
        return output


class Hdf5SingleFilePathProcessor(PathProcessor):
    """A Path processor for 3D datasets stored in a single hdf5 file.

    """
    def __init__(self):
        """Constructor to add custom class var"""
        PathProcessor.__init__(self)

    def setup(self, parameters):
        """Set the params

        MUST HAVE THE CUSTOM PARAMETERS: "filename": "<path_to_filet>",

        Includes the "ingest_job" section of the config file automatically

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        self.parameters = parameters

    def process(self, x_index, y_index, z_index, t_index=None):
        """
        Method to compute the file path for the indicated Z-slice

        Args:
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (str): An absolute file path that contains the specified data

        """
        # prepend root, append extension
        return self.parameters['filename']


class Hdf5SingleFileTileProcessor(TileProcessor):
    """A Tile processor for 3D datasets stored in a single HDF5 file

    Assumes the data is stored in a dataset and an optional offset is stored in a dataset

    """

    def __init__(self):
        """Constructor to add custom class var"""
        TileProcessor.__init__(self)
        self.fs = None

    def setup(self, parameters):
        """ Method to load the file for uploading

        Args:
            parameters (dict): Parameters for the dataset to be processed


        MUST HAVE THE CUSTOM PARAMETERS: "upload_format": "<png|tiff>",
                                         "data_name": str,
                                         "datatype": <uint8|uint16|uint32>
                                         "offset_x": int,
                                         "offset_y": int,
                                         "offset_z": int,
                                         "filesystem": "<s3|local>",
                                         "bucket": (if s3 filesystem)

        Returns:
            None
        """
        self.parameters = parameters
        self.fs = DynamicFilesystemAbsPath(parameters['filesystem'], parameters)

    def process(self, file_path, x_index, y_index, z_index, t_index=0):
        """
        Method to load the image file.

        Args:
            file_path(str): An absolute file path for the specified tile
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (io.BufferedReader): A file handle for the specified tile

        """
        file_path = self.fs.get_file(file_path)

        # Compute global range
        target_x_range = [self.parameters["ingest_job"]["tile_size"]["x"] * x_index,
                          self.parameters["ingest_job"]["tile_size"]["x"] * (x_index + 1)]
        target_y_range = [self.parameters["ingest_job"]["tile_size"]["y"] * y_index,
                          self.parameters["ingest_job"]["tile_size"]["y"] * (y_index + 1)]

        # Open hdf5
        h5_file = h5py.File(file_path, 'r')

        # Compute range in actual data, taking offsets into account
        x_offset = self.parameters['offset_x']
        y_offset = self.parameters['offset_y']
        x_tile_size = self.parameters["ingest_job"]["tile_size"]["x"]
        y_tile_size = self.parameters["ingest_job"]["tile_size"]["y"]

        h5_x_range = [target_x_range[0] + x_offset, target_x_range[1] + x_offset]
        h5_y_range = [target_y_range[0] + y_offset, target_y_range[1] + y_offset]
        h5_z_slice = z_index + self.parameters['offset_z']

        tile_x_range = [0, x_tile_size]
        tile_y_range = [0, y_tile_size]

        h5_max_x = h5_file[self.parameters['data_name']].shape[2]
        h5_max_y = h5_file[self.parameters['data_name']].shape[1]

        if h5_x_range[0] < 0:
            # insert sub-region into tile
            tile_x_range = [h5_x_range[0] * -1, x_tile_size]
            h5_x_range[0] = 0
        if h5_y_range[0] < 0:
            # insert sub-region into tile
            tile_y_range = [h5_y_range[0] * -1, y_tile_size]
            h5_y_range[0] = 0

        if h5_x_range[1] > h5_max_x:
            # insert sub-region into tile
            tile_x_range = [0, x_tile_size - (h5_x_range[1] - h5_max_x)]
            h5_x_range[1] = h5_max_x
        if h5_y_range[1] > h5_max_y:
            # insert sub-region into tile
            tile_y_range = [0, y_tile_size - (h5_y_range[1] - h5_max_y)]
            h5_y_range[1] = h5_max_y

        if self.parameters['datatype'] == "uint8":
            datatype = np.uint8
        elif self.parameters['datatype']== "uint16":
            datatype = np.uint16
        elif self.parameters['datatype']== "uint32":
            datatype = np.uint32
        else:
            raise Exception("Unsupported datatype: {}".format(self.parameters['datatype']))

        # Allocate Tile
        tile_data = np.zeros((self.parameters["ingest_job"]["tile_size"]["y"],
                             self.parameters["ingest_job"]["tile_size"]["x"]),
                             dtype=datatype, order='C')

        if h5_z_slice >= 0:
            # Copy sub-img to tile, save, return
            tile_data[tile_y_range[0]:tile_y_range[1],
                      tile_x_range[0]:tile_x_range[1]] = np.array(h5_file[self.parameters['data_name']][
                                                                          h5_z_slice,
                                                                          h5_y_range[0]:h5_y_range[1],
                                                                          h5_x_range[0]:h5_x_range[1]])

        tile_data = tile_data.astype(datatype)
        upload_img = Image.fromarray(tile_data)

        output = six.BytesIO()
        upload_img.save(output, format=self.parameters["upload_format"].upper())

        # Send handle back
        return output



# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
import six
from PIL import Image
import numpy as np
import os

from .path import PathProcessor
from .tile import TileProcessor


class CatmaidFileImageStackZoomLevelPathProcessor(PathProcessor):
    """Class for Catmaid File-based image stacks with zoom levels, Tile source type 4 in the documentation"""
    def setup(self, parameters):
        """Set the params

        MUST HAVE THE CUSTOM PARAMETERS: "root_dir": "<path_to_stack_root>",
                                         "filetype": "<png|tif|jpg>"
        Includes the "ingest_job" section of the config file automatically

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        self.parameters = parameters

    def process(self, x_index, y_index, z_index, t_index=None):
        """
        Method to compute the file path for the indicated tile - For this, it's always the same file for each Z slice

        Args:
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (str): An absolute file path that contains the specified data

        """
        if t_index != 0:
            raise IndexError("CATMAID File Image Stack format does not support non-zero time index")

        if z_index < self.parameters["ingest_job"]["extent"]["z"][0] or z_index >= self.parameters["ingest_job"]["extent"]["z"][1]:
            raise IndexError("Invalid Tile Z-Index: {} Z-Extent: {}".format(z_index, self.parameters["ingest_job"]["extent"]["z"][1]))

        if x_index > self.parameters["ingest_job"]["extent"]["x"][1] / self.parameters["ingest_job"]["tile_size"]["x"] - 1:
            raise IndexError("Invalid Tile X-Index: {} X-Extent: {} X-TileSize: {}".format(x_index,
                                                                                           self.parameters["ingest_job"]["extent"]["x"][1],
                                                                                           self.parameters["ingest_job"]["tile_size"]["x"]))

        if y_index > self.parameters["ingest_job"]["extent"]["y"][1] / self.parameters["ingest_job"]["tile_size"]["y"] - 1:
            raise IndexError("Invalid Tile Y-Index: {} Y-Extent: {} Y-TileSize: {}".format(y_index,
                                                                                           self.parameters["ingest_job"]["extent"]["y"][1],
                                                                                           self.parameters["ingest_job"]["tile_size"]["y"]))

        filename = "{}_{}.{}".format(y_index, x_index, self.parameters["filetype"])
        return os.path.join(self.parameters["root_dir"], "{}".format(self.parameters["ingest_job"]["resolution"]), "{}".format(z_index), filename)


class CatmaidFileImageStackZoomLevelTileProcessor(TileProcessor):
    """A Tile processor for a file where a multi-page TIFF contains all time points for a single z-slice"""
    def __init__(self):
        """Constructor to add custom class var"""
        TileProcessor.__init__(self)
        self.data = None

    def setup(self, parameters):
        """ Method to load the file for uploading - a very naive approach

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        self.parameters = parameters

    def process(self, file_path, x_index, y_index, z_index, t_index=0):
        """
        Method to load the configuration file and select the correct validator and backend

        Args:
            file_path(str): An absolute file path for the specified tile
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (io.BufferedReader): A file handle for the specified tile

        """
        # Save img to png and return handle
        tile_data = Image.open(file_path)

        output = six.BytesIO()
        tile_data.save(output, format=self.parameters["filetype"].upper())

        # Send handle back
        return output

class CatmaidDirectoryImageStackPathProcessor(PathProcessor):
    """Class for Catmaid Directory-based image stacks, Tile source type 5 in the documentation"""
    def setup(self, parameters):
        """Set the params

        MUST HAVE THE CUSTOM PARAMETERS: "root_dir": "<path_to_stack_root>",
                                         "filetype": "<png|tif|jpg>"
        Includes the "ingest_job" section of the config file automatically

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        self.parameters = parameters

    def process(self, x_index, y_index, z_index, t_index=None):
        """
        Method to compute the file path for the indicated tile - For this, it's always the same file for each Z slice

        Args:
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (str): An absolute file path that contains the specified data

        """
        if t_index != 0:
            raise IndexError("CATMAID File Image Stack format does not support non-zero time index")

        if z_index < self.parameters["ingest_job"]["extent"]["z"][0] or z_index >= self.parameters["ingest_job"]["extent"]["z"][1]:
            raise IndexError("Invalid Tile Z-Index: {}".format(z_index))

        if x_index > self.parameters["ingest_job"]["extent"]["x"][1] / self.parameters["ingest_job"]["tile_size"]["x"] - 1:
            raise IndexError("Invalid Tile X-Index: {}".format(x_index))

        if y_index > self.parameters["ingest_job"]["extent"]["y"][1] / self.parameters["ingest_job"]["tile_size"]["y"] - 1:
            raise IndexError("Invalid Tile Y-Index: {}".format(y_index))

        filename = "{}.{}".format(y_index, x_index, self.parameters["filetype"])
        return os.path.join(self.parameters["root_dir"], "{}".format(self.parameters["ingest_job"]["resolution"]), "{}".format(z_index), "{}".format(y_index), filename)


class CatmaidDirectoryImageStackTileProcessor(TileProcessor):
    """A Tile processor for a file where a multi-page TIFF contains all time points for a single z-slice"""
    def __init__(self):
        """Constructor to add custom class var"""
        TileProcessor.__init__(self)
        self.data = None

    def setup(self, parameters):
        """ Method to load the file for uploading - a very naive approach

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        self.parameters = parameters

    def process(self, file_path, x_index, y_index, z_index, t_index=0):
        """
        Method to load the configuration file and select the correct validator and backend

        Args:
            file_path(str): An absolute file path for the specified tile
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (io.BufferedReader): A file handle for the specified tile

        """
        # Save img to png and return handle
        tile_data = Image.open(file_path)

        output = six.BytesIO()
        tile_data.save(output, format=self.parameters["filetype"].upper())

        # Send handle back
        return output

class CatmaidFileImageStackPathProcessor(PathProcessor):
    """Class for Catmaid File-based image stacks, Tile source type 1 in the documentation"""
    def setup(self, parameters):
        """Set the params

        MUST HAVE THE CUSTOM PARAMETERS: "root_dir": "<path_to_stack_root>",
                                         "filetype": "<png|tif|jpg>"
        Includes the "ingest_job" section of the config file automatically

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        self.parameters = parameters

    def process(self, x_index, y_index, z_index, t_index=None):
        """
        Method to compute the file path for the indicated tile - For this, it's always the same file for each Z slice

        Args:
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (str): An absolute file path that contains the specified data

        """
        if t_index != 0:
            raise IndexError("CATMAID File Image Stack format does not support non-zero time index")

        if z_index < self.parameters["ingest_job"]["extent"]["z"][0] or z_index >= self.parameters["ingest_job"]["extent"]["z"][1]:
            raise IndexError("Invalid Tile Z-Index: {}".format(z_index))

        if x_index > self.parameters["ingest_job"]["extent"]["x"][1] / self.parameters["ingest_job"]["tile_size"]["x"] - 1:
            raise IndexError("Invalid Tile X-Index: {}".format(x_index))

        if y_index > self.parameters["ingest_job"]["extent"]["y"][1] / self.parameters["ingest_job"]["tile_size"]["y"] - 1:
            raise IndexError("Invalid Tile Y-Index: {}".format(y_index))

        filename = "{}_{}_{}.{}".format(y_index, x_index, self.parameters["ingest_job"]["resolution"], self.parameters["filetype"])
        return os.path.join(self.parameters["root_dir"], "{}".format(z_index), filename)


class CatmaidFileImageStackTileProcessor(TileProcessor):
    """A Tile processor for a file where a multi-page TIFF contains all time points for a single z-slice"""
    def __init__(self):
        """Constructor to add custom class var"""
        TileProcessor.__init__(self)
        self.data = None

    def setup(self, parameters):
        """ Method to load the file for uploading - a very naive approach

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        self.parameters = parameters

    def process(self, file_path, x_index, y_index, z_index, t_index=0):
        """
        Method to load the configuration file and select the correct validator and backend

        Args:
            file_path(str): An absolute file path for the specified tile
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (io.BufferedReader): A file handle for the specified tile

        """
        # Save img to png and return handle
        tile_data = Image.open(file_path)

        output = six.BytesIO()
        tile_data.save(output, format=self.parameters["filetype"].upper())

        # Send handle back
        return output



#!/usr/bin/env python
# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is a convenience script that ensures that the specified Boss resources
(collection, experiment, channel, and coordinate frame) exist before using
boss-ingest.  It takes a JSON input file that describes each Boss resource.
Note that if the Boss resource already exists, it will not be modified.  For
example, if the coordinate frame already exists, the properties specified in
the JSON input file will be ignored.  If the Boss resource does not exist, it
is created according to the specification in the input file.

The boss-ingest config file that is referenced by this script's JSON input
file is optionally updated with the name of the collection, experiment, and
channel.  This property is "ingest_cfg" and is only required when using the
`--writecfg` option.

Below is a sample JSON input file for this script.  If a property is not marked
with an "Optional" comment, it must be included in the JSON file.  In the 
sample JSON, below, the values assigned to the optional properties are the
default values that will be used if you do not provide these properties.  The
exception is "ingest_cfg" as explained previously.

{
  "ingest_cfg": "ingest.json",      # boss-ingest config associated with this.
  "collection": {
    "name": "my_collection",
    "description": "Example collection"
  },
  "coordinate_frame": {
    "name": "my_coord",
    "description": "Example coord frame",
    "x_start": 0,
    "x_stop": 1024,
    "y_start": 0,
    "y_stop": 1024,
    "z_start": 0,
    "z_stop": 32,
    "voxel_size_x": 4,
    "voxel_size_y": 4,
    "voxel_size_z": 40
  },
  "experiment": {
    "name": "my_exp",
    "description": "Example experiment",
    "num_time_samples": 1,                  # Optional.
    "num_hierarchy_levels": 1,              # Optional.
    "hierarchy_method": "anisotropic",      # Optional.
    "time_step": 0,                         # Optional.
    "time_step_unit": "nanoseconds"         # Optional.
  },
  "channel": {
    "name": "my_chan",
    "description": "Example channel",
    "type": "image",
    "datatype": "uint8",
    "sources": []                           # Optional names of source channels.
  }
}
"""

import argparse
from intern.remote.boss import BossRemote
from intern.resource.boss.resource import *
import json
import os


# Name of group to assign read-only privileges to.
GOV_TEAM = 'gov_team'

def create_collection(config, rmt):
    """
    Ensure collection exists.

    Args:
        config (dict):
        rmt (BossRemote):

    Returns:
        (CollectionResource)
    """
    collection = CollectionResource(
        config['collection']['name'], config['collection']['description'])
    try:
        collection = rmt.create_project(collection)
    except Exception as e:
        collection = rmt.get_project(collection)
    return collection


def create_experiment(config, rmt, collection, coord):
    """
    Ensure channel exists.

    Args:
        config (dict):
        rmt (BossRemote):
        collection (CollectionResource):
        coord (CoordinateFrameResource):

    Returns:
        (ExperimentResource)
    """

    # Default values if not provided in config file.
    exp_vals = {
        'hierarchy_method': 'anisotropic',
        'num_hierarchy_levels': 1,
        'num_time_samples': 1,
        'time_step': 0,
        'time_step_unit': 'seconds'
    }

    exp_vals.update(config['experiment'])

    experiment = ExperimentResource(
        config['experiment']['name'], collection.name, coord.name,
        config['experiment']['description'],
        num_hierarchy_levels=exp_vals['num_hierarchy_levels'],
        hierarchy_method=exp_vals['hierarchy_method'],
        num_time_samples=exp_vals['num_time_samples'],
        time_step=exp_vals['time_step'],
        time_step_unit=exp_vals['time_step_unit'])
    try:
        experiment = rmt.create_project(experiment)
    except:
        experiment = rmt.get_project(experiment)
    return experiment


def create_channel(config, rmt, collection, experiment):
    """
    Ensure channel exists.

    Args:
        config (dict):
        rmt (BossRemote):
        collection (CollectionResource):
        experiment (ExperimentResource):

    Returns:
        (ChannelResource)
    """
    if 'sources' not in config['channel']:
        sources = []
    else:
        sources = config['channel']['sources']

    channel = ChannelResource(
        config['channel']['name'], collection.name, experiment.name,
        type=config['channel']['type'], description=config['channel']['description'],
        datatype=config['channel']['datatype'], sources=sources)
    try:
        channel = rmt.create_project(channel)
    except:
        channel = rmt.get_project(channel)
    return channel


def create_coord_frame(config, rmt):
    """
    Ensure coordinate frame exists.

    Args:
        config (dict):
        rmt (BossRemote):

    Returns:
        (CoordinateFrameResource)
    """
    coord = CoordinateFrameResource(config['coordinate_frame']['name'],
                                    config['coordinate_frame']['description'],
                                    config['coordinate_frame']['x_start'],
                                    config['coordinate_frame']['x_stop'],
                                    config['coordinate_frame']['y_start'],
                                    config['coordinate_frame']['y_stop'],
                                    config['coordinate_frame']['z_start'],
                                    config['coordinate_frame']['z_stop'],
                                    config['coordinate_frame']['voxel_size_x'],
                                    config['coordinate_frame']['voxel_size_y'],
                                    config['coordinate_frame']['voxel_size_z'])
    try:
        coord = rmt.create_project(coord)
    except:
        coord = rmt.get_project(coord)
    return coord


def add_gov_team_permissions(rmt, collection, experiment, channel):
    """
    Give read permission to the government team.

    Args:
        rmt (BossRemote):
        collection (CollectionResource):
        experiment (ExperimentResource):
        channel (ChannelResource):
    """

    try:
        rmt.add_permissions(GOV_TEAM, collection, ['read'])
    except:
        print('Failed to automatically add {} group to collection: {}'.format(GOV_TEAM, collection.name))

    try:
        rmt.add_permissions(GOV_TEAM, experiment, ['read'])
    except:
        print('Failed to automatically add {} group to experiment: {}'.format(GOV_TEAM, experiment.name))
    try:
        rmt.add_permissions(GOV_TEAM, channel, ['read'])
    except:
        print('Failed to automatically add {} group to channel: {}'.format(GOV_TEAM, channel.name))


def main(parser_args=None):
    """
    Main entry point of script.

    Args:
        parser_args(argparse.ArgumentParser): A pre-loaded ArgumentParser instance
    """
    if parser_args is None:
        args = parse_args()
    else:
        args = parser_args

    rmt = BossRemote(args.intern_cfg)

    # Load Configuration File
    with open(args.config_file, 'rt') as cfg:
        config = json.load(cfg)

    collection = create_collection(config, rmt)
    coord = create_coord_frame(config, rmt)
    experiment = create_experiment(config, rmt, collection, coord)
    channel = create_channel(config, rmt, collection, experiment)

    if args.govteam:
        add_gov_team_permissions(rmt, collection, experiment, channel)

    # Update boss-ingest config file with resources names from config file
    if args.writecfg:
        with open(config['ingest_cfg'], 'rt') as cfg:
            ingest_file = json.load(cfg)
            if 'database' not in ingest_file:
                ingest_file['database'] = {}
            ingest_file['database']['collection'] = collection.name
            ingest_file['database']['experiment'] = experiment.name
            ingest_file['database']['channel'] = channel.name

        with open(config['ingest_cfg'], 'wt') as cfg:
            json.dump(ingest_file, cfg, indent=2)

    print('\n\nRun this command in the ingest-client repo directory to execute the ingest client:')
    print('\n  export INTERN_TOKEN={}'.format(rmt.project_service.auth))
    print('  boss-ingest {}'.format(os.path.abspath(config['ingest_cfg'])))


def parse_args():
    """
    Parse command line arguments.

    Returns:
        (Namespace)
    """
    parser = argparse.ArgumentParser(
        description='Script that creates the collection/experiment/channel for ingest.  ' + 
        'To supply arguments from a file, provide the filename prepended with an `@`.',
        fromfile_prefix_chars = '@')
    parser.add_argument(
        '--intern_cfg', '-i',
        default='boss.cfg',
        help='intern config file')
    parser.add_argument(
        '--writecfg',
        action='store_true',
        help='Update the boss-ingest config file with names of collection/experiment/channel.')
    parser.add_argument(
        '--govteam',
        action='store_true',
        help='Add read-only permissions for the Government team.')
    parser.add_argument(
        'config_file',
        help='Collection/experiment/channel configuration in JSON.')
    return parser.parse_args()


if __name__ == '__main__':
    main()




# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from abc import ABCMeta, abstractmethod

import six
from six.moves import cPickle as pickle
import importlib
from pkg_resources import resource_filename
import os

from .validator import Validator
from .backend import Backend


@six.python_2_unicode_compatible
class ConfigPropertyObject(object):
    def __init__(self, name, data=None, help_str=None, description=None):
        """

        Args:
            name(str): Name of the property object. Will be come the Key of the object when nested in a dict
            data(dict): Dictionary of properties to initialize with
            help_str(dict(str)): Dictionary of help strings that map to keys in data
            description(str): A description of this instance
        """
        if data:
            if not isinstance(data, dict):
                raise ValueError("'data' must be a dictionary of properties to add")
            self.__dict__ = data

        if help_str:
            if not isinstance(data, dict):
                raise ValueError("'help_str' must be a dictionary of help strings which correspond to 'data' keys")
            self.__help_str = help_str

        self.__object_name = name
        self.__object_description = description

    def __iter__(self):
        for x in self.get_properties():
            yield self.__dict__[x]

    def __str__(self):
        return self.__object_name

    def add_property(self, prop_name, prop_value, help_text=None):
        """
        Method to add a property

        Args:
            prop_name(str): name of the property
            prop_value: property value. can be anything that can be a dictionary value
            help_text(str): string of help information

        Returns:
            None
        """
        self.__dict__[prop_name] = prop_value

        if help_text:
            self.__help_str[prop_name] = help_text

    def add_property_object(self, prop_obj, help_text=None):
        """
        Method to add a property

        Args:
            prop_obj(ConfigPropertyObject): The property object to add
            help_text(str): String containing help information

        Returns:
            None
        """
        self.__dict__[prop_obj.get_name()] = prop_obj

        if help_text:
            self.__help_str[prop_obj.get_name()] = help_text

    def get_name(self):
        """Method to return the object name since we are mangling the object name so it doesn't get encoded"""
        return self.__object_name

    def get_description(self):
        """Method to return the object description since we are mangling the object name so it doesn't get encoded"""
        return self.__object_description

    def get_properties(self):
        """Method to return a list of property names contained in object

        Returns:
            (list(str)): List of property names
        """
        props = []
        for prop in [x for x in self.__dict__ if "__" not in x]:
            props.append(prop)

        return props

    def get_help_str(self, property_name):
        """Method to return the help string for a specific property

        Args:
            property_name(str): The name of the property

        Returns:
            (str): Help string
        """
        return self.__help_str[property_name]

    def explain(self):
        """Method to print help information for this configuration object

        Returns:
            None
        """
        print("\n\n### {} ###".format(self.__object_name))
        if self.__object_description:
            print(self.__object_description)
            print("")
        for prop in self.get_properties():
            print(" - {}: {}".format(prop, self.get_help_str(prop)))
        print("\n")

    def to_dict(self):
        """Method to recursively encode as a dictionary."""
        output = {}
        for prop in [x for x in self.__dict__ if "__" not in x]:
            if isinstance(self.__dict__[prop], ConfigPropertyObject):
                output.update(self.__dict__[prop].to_dict())
            else:
                # An actual property
                output[prop] = self.__dict__[prop]

        return {self.__object_name: output}


@six.add_metaclass(ABCMeta)
class ConfigurationGenerator(object):
    def __init__(self):
        self.config = ConfigPropertyObject("ROOT")
        self.name = "Base Config"
        self.description = ""

        # Setup the config
        self.setup()

    def load(self, file_path):
        """Method to pre-populate the class

        Args:
            file_path(str): Absolute path to the saved instance

        Returns:
            None
        """
        with open(file_path, 'rb') as file_handle:
            self.__dict__ = pickle.load(file_handle)

    def save(self, file_path):
        """

        Args:
            file_path(str): Absolute path to the file for saving a copy of the instance

        Returns:
            None
        """
        with open(file_path, 'wb') as file_handle:
            if six.PY3:
                pickle.dump(self.__dict__, file_handle, 2, fix_imports=True)
            else:
                pickle.dump(self.__dict__, file_handle, 2)

    @abstractmethod
    def setup(self):
        """Method to setup in instance by populating the correct data/property objects"""
        raise NotImplemented

    def to_json(self, file_path):
        """Method to serialize to json

        Args:
            file_path(str): Absolute path to the file for saving the instance as JSON

        Returns:
            None
        """
        config_dict = {}
        for obj in self.config:
            config_dict.update(obj.to_dict())

        with open(file_path, 'wt') as file_handle:
            json.dump(config_dict, file_handle)

    def explain(self):
        """Method to print help information for creating configuration

        Returns:
            None
        """
        print("\n\n### {} ###".format(self.name))
        if self.description:
            print(self.description)
            print("")
        for prop in self.config:
            print(" - {}: {}".format(prop.get_name(), prop.get_description()))
        print("\n")

    def add_property_object(self, prop_obj, help_text=None):
        """
        Method to add a property

        Args:
            prop_obj(ConfigPropertyObject): The property object to add
            help_text(str): String containing help information

        Returns:
            None
        """
        if not isinstance(prop_obj, ConfigPropertyObject):
            raise ValueError("Must add ConfigPropertyObject instance type")

        self.config.add_property_object(prop_obj, help_text=help_text)


class BossConfigurationGenerator(ConfigurationGenerator):
    def __init__(self):
        ConfigurationGenerator.__init__(self)

        self.name = "Boss Ingest v0.1"
        self.description = "Configuration Generator for the Boss ingest service v0.1"

    def setup(self):
        """Method to setup instance by populating the correct data/property objects"""
        # Schema section
        schema = ConfigPropertyObject("schema",
                                      {"version": "0.1",
                                       "name": "boss",
                                       "validator": "BossValidatorV01"},
                                      {"version": "Ingest service version number",
                                       "name": "Ingest service type",
                                       "validator": "The validator class to use to validate the schema"},
                                      "Schema properties for validation"
                                      )

        # Client Section
        client_backend = ConfigPropertyObject("backend",
                                              {"name": "boss",
                                               "class": "BossBackend",
                                               "host": "api.theboss.io",
                                               "protocol": "https"},
                                              {"name": "the backend name",
                                               "class": "the class to load for handling this backend",
                                               "host": "the domain name for the ingest server",
                                               "protocol": "https or http"},
                                              "Properties for the backend service"
                                              )
        client_tile_processor = ConfigPropertyObject("tile_processor",
                                                     {"class": "",
                                                      "params": None},
                                                     {"class": "The name of the class to load for tile processing",
                                                      "params": "Custom properties in a dictionary"},
                                                     "Properties for the custom tile processor class"
                                                     )
        client_file_processor = ConfigPropertyObject("file_processor",
                                                     {"class": "",
                                                      "params": None},
                                                     {"class": "The name of the class to load for file processing",
                                                      "params": "Custom properties in a dictionary"},
                                                     "Properties for the custom file processor class"
                                                     )

        client = ConfigPropertyObject("client",
                                      {"backend": client_backend,
                                       "tile_processor": client_tile_processor,
                                       "file_processor": client_file_processor},
                                      {"backend": "Properties for the backend service",
                                       "tile_processor": "Ingest service type",
                                       "file_processor": "The validator class to use to validate the schema"},
                                      "Ingest client properties"
                                      )

        database = ConfigPropertyObject("database",
                                        {"collection": "",
                                         "experiment": "",
                                         "channel": ""},
                                        {"collection": "The collection name containing the experiment and channel where data will be written",
                                         "experiment": "The experiment name containing the channel where data will be written",
                                         "channel": "The channel name where data will be written"},
                                        "Properties describing where in the database data should be written"
                                        )

        # Ingest Job Section
        ingest_job_extent = ConfigPropertyObject("extent",
                                                 {"x": [0, 1],
                                                  "y": [0, 1],
                                                  "z": [0, 1],
                                                  "t": [0, 1]},
                                                 {"x": "The spatial extent (in pixels) in the x-dimension. Python convention - start inclusive, stop exclusive",
                                                  "y": "The spatial extent (in pixels) in the y-dimension. Python convention - start inclusive, stop exclusive",
                                                  "z": "The spatial extent (in pixels) in the z-dimension. Python convention - start inclusive, stop exclusive",
                                                  "t": "The spatial extent (in pixels) in the t-dimension. Python convention - start inclusive, stop exclusive"},
                                                 "The spatial extent of the data to be uploaded.  This does not have to be the entire dataset."
                                                 )
        ingest_job_tile_size = ConfigPropertyObject("tile_size",
                                                 {"x": 4096,
                                                  "y": 4096,
                                                  "z": 1,
                                                  "t": 0},
                                                 {"x": "The size (in pixels) in the x-dimension of an individual image file",
                                                  "y": "The size (in pixels) in the y-dimension of an individual image file",
                                                  "z": "The size (in pixels) in the z-dimension of an individual image file",
                                                  "t": "The number of time samples in a single file"},
                                                 "The dimensions of the individual image data files that will be uploaded to the Boss"
                                                 )
        ingest_job = ConfigPropertyObject("ingest_job",
                                          {"resolution": 0,
                                           "extent": ingest_job_extent,
                                           "tile_size": ingest_job_tile_size},
                                          {"extent": "The spatial extent of the data to be uploaded.  This does not have to be the entire dataset.",
                                           "resolution": "The resolution level to ingest data. Default is 0 which represents native resolution",
                                           "tile_size": "The dimensions of the individual image data files that will be uploaded to the Boss"},
                                          "The properties defining what part of the dataset will be ingested"
                                          )

        self.add_property_object(schema)
        self.add_property_object(client)
        self.add_property_object(database)
        self.add_property_object(ingest_job)


class ConfigFileError(Exception):
    """Custom error to catch config file errors upstream in the client"""
    pass


class Configuration(object):
    def __init__(self, config_data=None):
        """
        A class to store configuration information and parameters for an ingest job

        Args:
            config_data(dict): The configuration dictionary
        """
        self.config_data = None
        self.schema = None
        self.tile_processor_class = None
        self.path_processor_class = None

        # If a configuration file was provided, load it now
        if config_data:
            self.load(config_data)

    def load(self, config_data):
        """
        Method to load the configuration file, the configuration schema, and select the correct validator and backend

        Args:
            config_data(dict): The configuration dictionary

        Returns:
            None

        """
        self.config_data = config_data

        # Load the schema file based on the config that was provided
        try:
            schema_name = self.config_data['schema']['name']
        except KeyError as err:
            raise ConfigFileError("The specified schema was not found: {}. Try to update your ingest client library or double check your ingest job configuration file".format(self.config_data['schema']['name']))
        with open(os.path.join(resource_filename("ingestclient", "schema"), "{}.json".format(schema_name)), 'rt') as schema_file:
            self.schema = json.load(schema_file)

    def load_plugins(self):
        """Method to load the plugins

        Returns:
            None
        """
        # Create plugin instances
        if "tile_processor" in self.config_data["client"]:
            package, class_name = self.config_data["client"]["tile_processor"]["class"].rsplit('.', 1)
            tile_module = importlib.import_module(package)
            tile_class = getattr(tile_module, class_name)
            self.tile_processor_class = tile_class()

        if "chunk_processor" in self.config_data["client"]:
            package, class_name = self.config_data["client"]["chunk_processor"]["class"].rsplit('.', 1)
            chunk_module = importlib.import_module(package)
            chunk_class = getattr(chunk_module, class_name)
            self.chunk_processor_class = chunk_class()

        package, class_name = self.config_data["client"]["path_processor"]["class"].rsplit('.', 1)
        path_module = importlib.import_module(package)
        path_class = getattr(path_module, class_name)
        self.path_processor_class = path_class()

    def get_chunk_processor_params(self):
        """Method to get the chunk processor parameter dictionary

        Returns:
            (dict): Dictionary of params from the config file
        """
        params = self.config_data["client"]["chunk_processor"]["params"]
        params["ingest_job"] = self.config_data["ingest_job"]
        return params

    def get_tile_processor_params(self):
        """Method to get the tile processor parameter dictionary

        Returns:
            (dict): Dictionary of params from the config file
        """
        params = self.config_data["client"]["tile_processor"]["params"]
        params["ingest_job"] = self.config_data["ingest_job"]
        return params

    def get_path_processor_params(self):
        """Method to get the path processor parameter dictionary

        Returns:
            (dict): Dictionary of params from the config file
        """
        params = self.config_data["client"]["path_processor"]["params"]
        params["ingest_job"] = self.config_data["ingest_job"]
        return params

    def get_validator(self):
        """
        Method to get a validator instance based on the configuration

        Returns:
            (ingestclient.core.validator.Validator): Validator instance

        """
        if not self.config_data:
            raise ValueError("No configuration file loaded.")

        # Setup Validator while sanitizing input
        if any(x in self.config_data["schema"]["validator"] for x in [";", ".", "import"]):
            raise ValueError("Schema Validator Class contains dangerous syntax. Please only list the Class Name.")
        else:
            return Validator.factory(self.config_data["schema"]["validator"], self.config_data)

    def get_backend(self, backend_api_token=None):
        """
        Method to get a backend instance based on the configuration

        Returns:
            (ingestclient.core.backend.Backend): Backend instance

        """
        if not self.config_data:
            raise ValueError("No configuration file loaded.")

        # Setup Backend while sanitizing input
        if any(x in self.config_data["client"]["backend"]["class"] for x in [";", ".", "import"]):
            raise ValueError("Backend Class contains dangerous syntax. Please only list the Class Name.")
        else:
            backend = Backend.factory(self.config_data["client"]["backend"]["class"],
                                      self.config_data)

        # Setup the backend
        backend.setup(backend_api_token)

        return backend

    def to_json(self):
        """
        Method to return a JSON string containing the configuration

        Returns:
            (str): JSON encoded config data
        """
        return json.dumps(self.config_data)




# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import six

from .path import PathProcessor
from .chunk import ChunkProcessor, XYZT_ORDER

from cloudvolume import CloudVolume

class CloudVolumePathProcessor(PathProcessor):
    """Class for simple image stacks that only increment in Z, uses the dynamic filesystem utility"""
    def __init__(self):
        """Constructor to add custom class var"""
        PathProcessor.__init__(self)

    def setup(self, parameters):
        """Set the params


        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        pass

    def process(self, x_index, y_index, z_index, t_index=None):
        """
        Method to compute the file path for the indicated tile

        Args:
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (str): An absolute file path that contains the specified data

        """
        return ""


class CloudVolumeChunkProcessor(ChunkProcessor):
    """Chunk processor that utilizes Seung Lab's CloudVolume"""

    def __init__(self):
        ChunkProcessor.__init__(self)
        self.vol = None
        self.ingest_job = None

    def setup(self, parameters):
        """ Method to load the file for uploading data. Assumes intern token is set via environment variable or config
        default file

        Args:
            parameters (dict): Parameters for the dataset to be processed


        MUST HAVE THE CUSTOM PARAMETER: "cloudpath": location of CloudVolume data

        All other parameters are optional parameters that may be passed to the CloudVolume constructor

        Returns:
            None
        """
        self.parameters = parameters

        # Remove 'ingest_job' key so rest of parameters can be passed to the
        # CloudVolume constructor.
        self.ingest_job = self.parameters.pop('ingest_job')
        self.vol = CloudVolume(**self.parameters)

    def process(self, file_path, x_index, y_index, z_index):
        """
        Method to take a chunk indices and return an ndarray with the correct data

        Args:
            file_path(str): An absolute file path for the specified chunk
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension

        Returns:
            (np.ndarray, int): ndarray for the specified chunk, XYZT_ORDER
        """
        x_size = self.ingest_job["chunk_size"]["x"]
        y_size = self.ingest_job["chunk_size"]["y"]
        z_size = self.ingest_job["chunk_size"]["z"]

        x_start = x_index * x_size;
        y_start = y_index * y_size;
        z_start = z_index * z_size;

        x_stop = x_start + x_size
        y_stop = y_start + y_size
        z_stop = z_start + z_size

        if x_stop > self.vol.bounds.maxpt[0]:
            x_stop = self.vol.bounds.maxpt[0]
        if y_stop > self.vol.bounds.maxpt[1]:
            y_stop = self.vol.bounds.maxpt[1]
        if z_stop > self.vol.bounds.maxpt[2]:
            z_stop = self.vol.bounds.maxpt[2]

        return self.vol[x_start:x_stop, y_start:y_stop, z_start:z_stop], XYZT_ORDER



# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
from abc import ABCMeta, abstractmethod
import requests
import json
import boto3
import hashlib
from six.moves import configparser
import time
import botocore
from pkg_resources import resource_filename
import os
import random

from ..utils import WaitPrinter
from ..utils.log import always_log_info
from ..utils.backoff import get_wait_time


@six.add_metaclass(ABCMeta)
class Backend(object):
    """

    Attributes:
        config (dict):
        sqs:
        s3 (boto3.S3)::
        upload_queue (boto3.SQS.Queue): Queue that holds upload tile messages.
        tile_index_queue (boto3.SQS.Queue): Queue that triggers a tile index update.
        bucket (S3.Bucket): Tile bucket.
        volumetric_bucket (S3.Bucket): Temporary cuboid holding bucket for volumetric ingests.
    """
    def __init__(self, config):
        """
        A class to implement a backend that supports the ingest service

        Args:
            config (dict): Dictionary of parameters from the "backend" section of the config file

        """
        self.config = config
        self.sqs = None
        self.upload_queue = None
        self.tile_index_queue = None
        self.s3 = None
        self.bucket = None
        self.volumetric_bucket = None

    @abstractmethod
    def setup(self):
        """
        Method to configure the backend based on configuration parameters in the config file

        Args:

        Returns:
            None


        """
        return NotImplemented

    @abstractmethod
    def create(self, data):
        """
        Method to upload the config data to the backend to create an ingest job

        Args:
            data(dict): A dictionary of configuration parameters

        Returns:
            (int): The returned ingest_job_id


        """
        return NotImplemented

    @abstractmethod
    def join(self, ingest_job_id):
        """
        Method to join an ingest job upload

        Job Status: {0: Preparing, 1: Uploading, 2: Complete, 3: Deleted}

        Args:
            ingest_job_id(int): The ID of the job you'd like to resume processing

        Returns:
            (int, dict, str, str, str, dict, int): The job status, AWS credentials, and SQS upload_job_queue,
                                              tile_index_queue, tile bucket name, config_params to pass along
                                              during upload via metadata, and tile count
        """
        return NotImplemented

    @abstractmethod
    def cancel(self, ingest_job_id):
        """
        Method to cancel an ingest job

        Args:
            ingest_job_id(int): The ID of the job you'd like to cancel

        Returns:
            None


        """
        return NotImplemented

    @abstractmethod
    def complete(self, ingest_job_id):
        """
        Method to complete an ingest job

        Args:
            ingest_job_id(int): The ID of the job you'd like to complete

        Returns:
            (bool): True on successful completion


        """
        return NotImplemented

    @abstractmethod
    def get_job_status(self, ingest_job_id):
        """
        Method to get the ingest job status

        Args:
            ingest_job_id(int): The ID of the job you'd like to resume processing

        Returns:
            (dict): Job status dictionary, including the number of messages in the queue


        """
        return NotImplemented

    @abstractmethod
    def get_task(self):
        """
        Method to get an upload task

        Args:

        Returns:
            None


        """
        return NotImplemented

    def setup_queues(self, credentials, upload_queue, tile_index_queue, region="us-east-1"):
        """
        Method to create a connection to the upload task queue

        Args:
            credentials(dict): AWS credentials
            upload_queue(str): The URL for the upload SQS queue
            tile_index_queue(str): URL of tile index queue
            region(str): The AWS region where the SQS queue exists

        Returns:
            None

        """
        self.sqs = boto3.resource('sqs', region_name=region, aws_access_key_id=credentials["access_key"],
                                  aws_secret_access_key=credentials["secret_key"])
        self.upload_queue = self.sqs.Queue(url=upload_queue)
        if tile_index_queue:
            self.tile_index_queue = self.sqs.Queue(url=tile_index_queue)

    def setup_tile_bucket(self, credentials, tile_bucket, region="us-east-1"):
        """
        Method to create a connection to the tile bucket

        Args:
            credentials(dict): AWS credentials
            tile_bucket(str): The name of the bucket
            region(str): The AWS region where the bucket exists

        Returns:
            None

        """
        self.s3 = boto3.resource('s3', region_name=region, aws_access_key_id=credentials["access_key"],
                                 aws_secret_access_key=credentials["secret_key"])
        self.bucket = self.s3.Bucket(tile_bucket)

    def setup_volumetric_bucket(self, credentials, bucket_name, region="us-east-1"):
        """
        Method to create a connection to the tile bucket

        Args:
            credentials(dict): AWS credentials
            bucket_name(str): The name of the bucket
            region(str): The AWS region where the bucket exists

        Returns:
            None

        """
        self.s3 = boto3.resource('s3', region_name=region, aws_access_key_id=credentials["access_key"],
                                 aws_secret_access_key=credentials["secret_key"])
        self.volumetric_bucket = self.s3.Bucket(bucket_name)

    @abstractmethod
    def encode_tile_key(self, project_info, resolution, x_index, y_index, z_index, t_index=0):
        """A method to create a tile key.

        The tile key is the key used for each individual tile file.

        Args:
            project_info(list): A list of strings containing the project/data model information for where data belongs
            resolution(int): The level of the resolution hierarchy.  Typically 0
            x_index(int): The x tile index
            y_index(int): The y tile index
            z_index(int): The z tile index
            t_index(int): The time index

        Returns:
            (str): The object key to use for uploading to the tile bucket
        """
        return NotImplemented

    @abstractmethod
    def encode_chunk_key(self, num_tiles, project_info, resolution, x_index, y_index, z_index, t_index=0):
        """A method to create a chunk key.

        A "chunk" is the group of tiles that must be uploaded so a cuboid can be ingested.  The chunk key is used
        to track all tiles in a given group.

        Args:
            num_tiles(int): The expected number of tiles in this chunk (in the z-dimension). Useful for forcing ingest of partial cuboids
            project_info(list): A list of strings containing the project/data model information for where data belongs
            resolution(int): The level of the resolution hierarchy.  Typically 0
            x_index(int): The x tile index
            y_index(int): The y tile index
            z_index(int): The z tile index
            t_index(int): The time index

        Returns:
            (str): The object key to use for uploading to the tile bucket
        """
        return NotImplemented

    @abstractmethod
    def decode_tile_key(self, key):
        """A method to decode the tile key

        The tile key is the key used for each individual tile file.

        Args:
            key(str): The key to decode

        Returns:
            (dict): A dictionary containing the components of the key
        """
        return NotImplemented

    @abstractmethod
    def decode_chunk_key(self, key):
        """A method to decode the chunk key

        The tile key is the key used for each individual tile file.

        Args:
            key(str): The key to decode

        Returns:
            (dict): A dictionary containing the components of the key
        """
        return NotImplemented

    @staticmethod
    def factory(backend_str, config_data):
        """
        Method to return a validator class based on a string
        Args:
            backend_str (str): String of the classname

        Returns:

        """
        if backend_str == "BossBackend":
            return BossBackend(config_data)
        else:
            return ValueError("Unsupported Backend: {}".format(backend_str))


class BossBackend(Backend):
    def __init__(self, config):
        """
        A class to implement a backend that supports the ingest service

        api_token is a dictionary with keys:

        Args:

        """
        self.host = None
        self.api_headers = None
        Backend.__init__(self, config)
        self.api_version = "latest"
        self.validate_ssl = True
        self.credential_timeout = 3300  # Currently credentials expire in 1 hr, so renew after 55 minutes

    def setup(self, api_token=None):
        """
        Method to configure the backend based on configuration parameters in the config file

        Args:

        Returns:
            None


        """
        self.host = "{}://{}".format(self.config["client"]["backend"]["protocol"],
                                     self.config["client"]["backend"]["host"])

        # If API token not provided, load API credentials from intern locations as needed.
        if not api_token:
            # Try environment var
            if "INTERN_TOKEN" in os.environ:
                api_token = os.environ["INTERN_TOKEN"]
            else:
                # Try to see if intern config file is setup
                try:
                    cfg_parser = configparser.ConfigParser()
                    cfg_parser.read(os.path.expanduser("~/.intern/intern.cfg"))
                    if "Default" in cfg_parser.sections():
                        api_token = cfg_parser.get("Default", "token")
                    elif "Project Service" in cfg_parser.sections():
                        api_token = cfg_parser.get("Project Service", "token")
                    else:
                        raise ValueError("Could not load config from ~/.intern/intern.cfg")
                except KeyError as e:
                    print("API Token not provided. Failed to setup backend: {}".format(e))

        self.api_headers = {'Authorization': 'Token ' + api_token, 'Accept': 'application/json',
                            'content-type': 'application/json'}

    def create(self, config_dict):
        """
        Method to upload the config data to the backend to create an ingest job

        Args:
            config_dict(dict): config data

        Returns:
            (int): The returned ingest_job_id


        """
        always_log_info("Submitting ingest job configuration for creation...")
        r = requests.post('{}/{}/ingest/'.format(self.host, self.api_version), json=config_dict,
                          headers=self.api_headers, verify=self.validate_ssl)

        if r.status_code != 201:
            msg = r.json()
            err_detail = None
            if "detail" in msg:
                err_detail = msg["detail"]
            elif "message" in msg:
                err_detail = msg["message"]

            raise Exception("Failed to create ingest job. Server side validation of configuration file failed: {}".format(err_detail))
        else:
            return r.json()['id']

    def join(self, ingest_job_id):
        """
        Method to join an ingest job upload

        Job Status: {0: Preparing, 1: Uploading, 2: Complete, 3: Deleted}

        Args:
            ingest_job_id(int): The ID of the job you'd like to resume processing

        Returns:
            (int, dict, str, str, str, dict, int): The job status, AWS credentials, and SQS upload_job_queue,
                                              tile_index_queue, tile bucket name, config_params to pass along
                                              during upload via metadata, and tile count
        """

        wp = WaitPrinter()
        maximum_retries = 100
        retries = 0
        while True:
            r = requests.get('{}/{}/ingest/{}'.format(self.host, self.api_version, ingest_job_id),
                             headers=self.api_headers, verify=self.validate_ssl)
            if r.status_code in [500, 502]:
                retries += 1
                if retries > maximum_retries:
                    raise Exception("After {} attempts, failed to join ingest job: {}".format(maximum_retries, r.text))

                exp_backoff = 100 * 2 ** retries  # in ms
                pause_for = random.uniform(1, min(30000, exp_backoff)) / 1000
                print("Join request failed with: {} pausing for {:0.3f} seconds before retrying.".format(r.status_code, pause_for))
                time.sleep(pause_for)
            elif r.status_code != 200:
                raise Exception("Failed to join ingest job: {}".format(r.text))
            else:
                result = r.json()
                job_status = int(result['ingest_job']["status"])
                wp.print_msg("(pid={}) Waiting for ingest job to be created".format(os.getpid()))
                if job_status == 0:
                    time.sleep(5)
                else:
                    wp.finished()
                    creds = result["credentials"]

                    # Add check to make sure valid credentials came back. If not, try again
                    if not creds:
                        continue

                    upload_queue = result["ingest_job"]["upload_queue"]
                    tile_index_queue = result["ingest_job"]["tile_index_queue"]
                    tile_bucket = result["tile_bucket_name"]
                    num_tiles = result["ingest_job"]["tile_count"]

                    # Setup params for the rest of the ingest process
                    # These params are passed as metadata that accompanies the
                    # S3 object.
                    params = {}
                    params["upload_queue"] = result["ingest_job"]["upload_queue"]
                    params["ingest_queue"] = result["ingest_job"]["ingest_queue"]
                    params["ingest_lambda"] = result["ingest_lambda"]
                    params["KVIO_SETTINGS"] = result["KVIO_SETTINGS"]
                    params["STATEIO_CONFIG"] = result["STATEIO_CONFIG"]
                    params["OBJECTIO_CONFIG"] = result["OBJECTIO_CONFIG"]
                    params["resource"] = result["resource"]

                    self.setup_queues(creds, upload_queue, tile_index_queue, region="us-east-1")
                    self.setup_tile_bucket(creds, tile_bucket, region="us-east-1")
                    if "ingest_bucket_name" in result:
                        self.setup_volumetric_bucket(creds, result["ingest_bucket_name"], region="us-east-1")

                    return job_status, creds, upload_queue, tile_index_queue, tile_bucket, params, num_tiles

    def cancel(self, ingest_job_id):
        """
        Method to cancel an ingest job

        Args:
            ingest_job_id(int): The ID of the job you'd like to resume processing

        Returns:
            None


        """
        r = requests.delete('{}/{}/ingest/{}'.format(self.host, self.api_version, ingest_job_id),
                            headers=self.api_headers, verify=self.validate_ssl)

        if r.status_code != 204:
            raise Exception("Failed to cancel ingest job: {}".format(r.json()))

    def complete(self, ingest_job_id):
        """
        Method to complete an ingest job

        Args:
            ingest_job_id(int): The ID of the job you'd like to complete

        Returns:
            (bool): True if ingest complete


        """
        r = requests.post('{}/{}/ingest/{}/complete'.format(self.host, self.api_version, ingest_job_id),
                          headers=self.api_headers, verify=self.validate_ssl)

        if r.status_code == 204:
            return True
        if r.status_code == 202:
            # Not all chunks ingested.
            return False

        raise Exception("Failed to complete ingest job: {}".format(r.json()))

    def verify(self, ingest_job_id):
        """
        Method to start verification of an ingest job

        Args:
            ingest_job_id(int): The ID of the job you'd like to verify

        Returns:
            (bool): True if verified, False if there are outstanding tiles


        """
        r = requests.post('{}/{}/ingest/{}/verify'.format(self.host, self.api_version, ingest_job_id),
                          headers=self.api_headers, verify=self.validate_ssl)

        if r.status_code == 204:
            return True
        
        # Verification process ran but still have work to do.
        if r.status_code == 202:
            return False

        # Something went wrong.
        raise Exception("Failed to verify ingest job: {}".format(r.json()))

    def get_task(self, num_messages=1):
        """
        Method to get an upload task

        Args:
            num_messages(int): Number of messages to pop off the upload task queue

        Returns:
            (str, str, dict): message_id, receipt_handle, message contents
        """
        try_cnt = 0
        while try_cnt < 19:
            try:
                msg = self.upload_queue.receive_messages(MaxNumberOfMessages=1, WaitTimeSeconds=1)
                break
            except botocore.exceptions.ClientError as e:
                print("(pid={}) Waiting for credentials to be valid".format(os.getpid()))
                try_cnt += 1
                time.sleep(15)

                if try_cnt >= 20:
                    raise Exception("(pid={}) Credentials failed to be come valid".format(os.getpid()))

        if msg:
            return msg[0].message_id, msg[0].receipt_handle, json.loads(msg[0].body)
        else:
            return None, None, None

    def delete_task(self, msg_id, receipt_handle):
        """
        Delete a message from the upload queue

        Args:
            msg_id (str): Id of queue message.
            receipt_handle (str): Actual id required to delete the message.

        Returns:
            (bool): True on success.

        Raises:
            (Exception): Raised after n consecutive ClientErrors.
        """
        MAX_TRIES = 20
        try_cnt = 0
        while try_cnt < MAX_TRIES - 1:
            try:
                resp = self.upload_queue.delete_messages(Entries=[{'Id': msg_id, 'ReceiptHandle': receipt_handle}])
                if 'Successful' in resp and len(resp['Successful']) > 0:
                    if resp['Successful'][0]['Id'] == msg_id:
                        return True

                # Failed for some reason.
                try_cnt += 1
                if 'Failed' in resp and len(resp['Failed']) > 0:
                    err = resp['Failed'][0]
                    always_log_info('Failed deleting message from queue: ({}) - {}'.format(err['Code'], err['Message']))
                    if err['SenderFault']:
                        # If it's our fault, give up.
                        break

                time.sleep(get_wait_time(try_cnt))
            except botocore.exceptions.ClientError:
                print("(pid={}) Waiting for credentials to be valid".format(os.getpid()))
                try_cnt += 1
                time.sleep(15)

                if try_cnt >= MAX_TRIES:
                    raise Exception("(pid={}) Credentials failed to be come valid".format(os.getpid()))

        return False

    def put_task(self, msg, max_retries):
        """
        Place the message from the upload queue on the tile index queue so the 
        tile index is updated.

        Args:
            msg (str): Message contents.
            max_retries (int): Max number retries to put message on queue.

        Returns:
            (bool): False on failure.
        """
        try_cnt = 0
        while try_cnt < max_retries + 1:
            try:
                self.tile_index_queue.send_message(MessageBody=msg)
                return True
            except botocore.exceptions.ClientError:
                try_cnt += 1
                time.sleep(get_wait_time(try_cnt))

        return False

    def get_job_status(self, ingest_job_id):
        """
        Method to get the job status

        Args:
            ingest_job_id(int): The ID of the job you'd like to resume processing

        Returns:
            (int)
        """

        maximum_retries = 100
        retries = 0
        while True:
            r = requests.get('{}/{}/ingest/{}/status'.format(self.host, self.api_version, ingest_job_id),
                             headers=self.api_headers, verify=self.validate_ssl)
            if r.status_code in [500, 502]:
                retries += 1
                if retries > maximum_retries:
                    raise Exception("After {} attempts, failed to join ingest job: {}".format(maximum_retries, r.text))

                exp_backoff = 100 * 2 ** retries  # in ms
                pause_for = random.uniform(1, min(30000, exp_backoff)) / 1000
                print("Job status request failed with: {} pausing for {:0.3f} seconds before retrying.".format(r.status_code,
                                                                                                         pause_for))
                time.sleep(pause_for)

            elif r.status_code != 200:
                raise Exception("Failed to get ingest job status: {}".format(r.text))
            else:
                return r.json()

    def encode_tile_key(self, project_info, resolution, x_index, y_index, z_index, t_index=0):
        """A method to create a tile key.

        The tile key is the key used for each individual tile file.

        Args:
            project_info(list): A list of strings containing the project/data model information for where data belongs
            resolution(int): The level of the resolution hierarchy.  Typically 0
            x_index(int): The x tile index
            y_index(int): The y tile index
            z_index(int): The z tile index
            t_index(int): The time index

        Returns:
            (str): The object key to use for uploading to the tile bucket
        """
        proj_str = six.u("&".join([str(x) for x in project_info]))
        base_key = six.u("{}&{}&{}&{}&{}&{}".format(proj_str, resolution, x_index, y_index, z_index, t_index))

        hashm = hashlib.md5()
        hashm.update(base_key.encode())

        return six.u("{}&{}".format(hashm.hexdigest(), base_key))

    def encode_chunk_key(self, num_tiles, project_info, resolution, x_index, y_index, z_index, t_index=0):
        """A method to create a chunk key.

        A "chunk" is either a single 3D volume or a group of tiles that must be uploaded so cuboids can be ingested.  The chunk key
        identifies the 3D volume or all tiles in a given group.

        Args:
            num_tiles(int): The expected number of tiles in this chunk (in the z-dimension). Useful for forcing ingest of partial cuboids.  For a 3D volume, its value is 1.
            project_info(list): A list of strings containing the project/data model information for where data belongs
            resolution(int): The level of the resolution hierarchy.  Typically 0
            x_index(int): The x tile index
            y_index(int): The y tile index
            z_index(int): The z tile index
            t_index(int): The time index

        Returns:
            (str): The object key to use for uploading to the tile bucket
        """
        proj_str = six.u("&".join([str(x) for x in project_info]))
        base_key = six.u("{}&{}&{}&{}&{}&{}&{}".format(num_tiles, proj_str,
                                                       resolution, x_index, y_index, z_index, t_index))

        hashm = hashlib.md5()
        hashm.update(base_key.encode())

        return six.u("{}&{}".format(hashm.hexdigest(), base_key))

    def decode_tile_key(self, key):
        """A method to decode the tile key

        The tile key is the key used for each individual tile file.

        Args:
            key(str): The key to decode

        Returns:
            (dict): A dictionary containing the components of the key
        """
        result = {}
        parts = key.split('&')
        result["collection"] = int(parts[1])
        result["experiment"] = int(parts[2])
        result["channel"] = int(parts[3])
        result["resolution"] = int(parts[4])
        result["x_index"] = int(parts[5])
        result["y_index"] = int(parts[6])
        result["z_index"] = int(parts[7])
        result["t_index"] = int(parts[8])

        return result

    def decode_chunk_key(self, key):
        """A method to decode the chunk key

        Args:
            key(str): The key to decode

        Returns:
            (dict): A dictionary containing the components of the key
        """
        result = {}
        parts = key.split('&')
        result["num_tiles"] = int(parts[1])
        result["collection"] = int(parts[2])
        result["experiment"] = int(parts[3])
        result["channel"] = int(parts[4])
        result["resolution"] = int(parts[5])
        result["x_index"] = int(parts[6])
        result["y_index"] = int(parts[7])
        result["z_index"] = int(parts[8])
        result["t_index"] = int(parts[9])

        return result



# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
from abc import ABCMeta, abstractmethod
import numpy as np

XYZ_ORDER = 0
ZYX_ORDER = 1
XYZT_ORDER = 2
TZYX_ORDER = 3

@six.add_metaclass(ABCMeta)
class ChunkProcessor(object):
    def __init__(self):
        """
        A class that implements a chunk processor which outputs ndarrays for uploading

        Args:
        """
        self.parameters = None

    @abstractmethod
    def setup(self, parameters):
        """
        Method to initialize the chunk processor based on custom parameters from the configuration file

        e.g. Connect to a database, etc.

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        return NotImplemented

    @abstractmethod
    def process(self, file_path, x_index, y_index, z_index):
        """
        Method to take a chunk indices and return an ndarray with the correct data

        Args:
            file_path(str): An absolute file path for the specified chunk
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension

        Returns:
            (np.ndarray, int): ndarray for the specified chunk, array order (XYZ_ORDER, TZYX_ORDER, etc)
        """
        return NotImplemented



# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta, abstractmethod
import six
import jsonschema
import json
from .consts import BOSS_CUBOID_X, BOSS_CUBOID_Y, BOSS_CUBOID_Z 


@six.add_metaclass(ABCMeta)
class Validator(object):
    def __init__(self, config_data):
        """
        A class to implement the ingest job configuration file validator

        Args:
            config_data(dict): Dictionary of configuration

        """
        self.config = config_data

        # Schema to be populated by backend
        self.schema = None

    def validate_schema(self):
        """
        Method to validate the JSON data against the schema

        Args:

        Returns:
            str: An error message if schema validation fails

        """
        if not self.schema:
            raise ValueError("Schema has not been populated yet. Cannot validate.")

        try:
            jsonschema.validate(self.config, self.schema)
        except jsonschema.ValidationError as e:
            return e

        return None

    def validate(self):
        """
        Method to load the configuration file and select the correct validator and backend

        Args:

        Returns:
            (dict): Dictionary of "info", "question", "error" messages

        """
        schema_err_msg = self.validate_schema()
        if not schema_err_msg:
            schema_info_msg = ["Configuration file schema validation - Passed"]
            info_msg, question_msg, error_msg = self.validate_properties()
            schema_info_msg.extend(info_msg)
            return {"info": schema_info_msg,
                    "question": question_msg,
                    "error": error_msg}
        else:
            return {"info": ["Configuration file schema validation - Failed"],
                    "question": [],
                    "error": [schema_err_msg.message]}

    @abstractmethod
    def validate_properties(self):
        """
        Method to validate any custom properties beyond verifying that the schema was used correctly

        Args:

        Returns:
            (list(str), list(str), list(str)): a tuple of lists containing "info", "question", "error" messages

        """
        return NotImplemented

    @staticmethod
    def factory(validator_str, config_data):
        """
        Method to return a validator class based on a string
        Args:
            validator_str(str): Class name for the validator to use
            config_data(dict): Dictionary containing configuration data

        Returns:

        """
        if validator_str == "BossValidatorV01":
            return BossValidatorV01(config_data)
        if validator_str == "BossValidatorV02":
            return BossValidatorV02(config_data)
        else:
            return ValueError("Unsupported validator: {}".format(validator_str))


class BossValidatorV01(Validator):
    def __init__(self, config_data):
        """
        A class to implement the ingest job configuration file validator for the Boss (docs.theBoss.io)

        Args:
            config_data(dict): Configuration dictionary

        """
        Validator.__init__(self, config_data)

    def validate_properties(self):
        """
        Method to validate any custom properties beyond verifying that the schema was used correctly

        Args:

        Returns:
            (list(str), list(str), list(str)): a tuple of lists containing "info", "question", "error" messages

        """
        # TODO: Add Boss specific validation
        # Verify Collection

        # Verify Experiment

        # Verify Channel

        # If channel already exists, check corners to see if data exists.  If so question user for overwrite

        # Check tile size - error if too big

        # Check backend connectivity

        return ['Parameter Validation Passed'], [], []


class BossValidatorV02(Validator):
    def __init__(self, config_data):
        """
        A class to implement the ingest job configuration file validator for the Boss (docs.theBoss.io)

        Args:
            config_data(dict): Configuration dictionary

        """
        Validator.__init__(self, config_data)

    def validate_properties(self):
        """
        Method to validate any custom properties beyond verifying that the schema was used correctly

        Args:

        Returns:
            (list(str), list(str), list(str)): a tuple of lists containing "info", "question", "error" messages

        """
        # TODO: Add Boss specific validation
        # Verify Collection

        # Verify Experiment

        # Verify Channel

        # If channel already exists, check corners to see if data exists.  If so question user for overwrite

        # Check tile size - error if too big

        # Check backend connectivity

        errors = []

        if self.config['ingest_job']['ingest_type'] == 'tile':
            if not 'tile_size' in self.config['ingest_job']:
                errors.append('"ingest_job": "tile_size" required for tile ingest jobs.')
            if not 'tile_processor' in self.config['client']:
                errors.append('"client": "tile_processor" required for tile ingest jobs.')
        elif self.config['ingest_job']['ingest_type'] == 'volumetric':
            if not 'chunk_size' in self.config['ingest_job']:
                errors.append('"ingest_job": "chunk_size" required for volumetric ingest jobs.')
            else:
                if self.config['ingest_job']['chunk_size']['x'] % BOSS_CUBOID_X != 0:
                    errors.append('"ingest_job": "chunk_size": "x" must be a multiple of {}.'.format(BOSS_CUBOID_X))
                if self.config['ingest_job']['chunk_size']['y'] % BOSS_CUBOID_Y != 0:
                    errors.append('"ingest_job": "chunk_size": "y" must be a multiple of {}.'.format(BOSS_CUBOID_Y))
                if self.config['ingest_job']['chunk_size']['z'] % BOSS_CUBOID_Z != 0:
                    errors.append('"ingest_job": "chunk_size": "z" must be a multiple of {}.'.format(BOSS_CUBOID_Z))
            if not 'chunk_processor' in self.config['client']:
                errors.append('"client": "chunk_processor" required for volumetric ingest jobs.')

        if len(errors) == 0:
            return ['Parameter Validation Passed'], [], errors

        return ["Configuration file schema validation - Failed"], [], errors



# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
import six
from PIL import Image
from intern.remote.boss import BossRemote
from intern.resource.boss.resource import ChannelResource
import numpy as np
import time

from .path import PathProcessor
from .tile import TileProcessor


class InternPathProcessor(PathProcessor):
    """Class for simple image stacks that only increment in Z, uses the dynamic filesystem utility"""
    def __init__(self):
        """Constructor to add custom class var"""
        PathProcessor.__init__(self)

    def setup(self, parameters):
        """Set the params


        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        pass

    def process(self, x_index, y_index, z_index, t_index=None):
        """
        Method to compute the file path for the indicated tile

        Args:
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (str): An absolute file path that contains the specified data

        """
        return ""


class InternTileProcessor(TileProcessor):
    """A Tile processor for a single image file identified by z index"""

    def __init__(self):
        """Constructor to add custom class var"""
        TileProcessor.__init__(self)
        self.remote = None
        self.channel = None

    def setup(self, parameters):
        """ Method to load the file for uploading data. Assumes intern token is set via environment variable or config
        default file

        Args:
            parameters (dict): Parameters for the dataset to be processed


        MUST HAVE THE CUSTOM PARAMETERS: "x_offset": offset to apply when querying the Boss
                                         "y_offset": offset to apply when querying the Boss
                                         "z_offset": offset to apply when querying the Boss
                                         "x_tile": size of a tile in x dimension
                                         "y_tile": size of a tile in y dimension
                                         "collection": source collection
                                         "experiment": source experiment
                                         "channel": source channel
                                         "resolution": source resolution

        Returns:
            None
        """
        self.parameters = parameters
        self.remote = BossRemote()
        self.channel = ChannelResource(self.parameters["channel"],
                                       self.parameters["collection"],
                                       self.parameters["experiment"])
        self.channel = self.remote.get_project(self.channel)

    def process(self, file_path, x_index, y_index, z_index, t_index=0):
        """
        Method to load the image file.

        Args:
            file_path(str): An absolute file path for the specified tile
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (io.BufferedReader): A file handle for the specified tile

        """
        # Compute cutout args
        x_rng = [self.parameters["x_tile"] * x_index + self.parameters["x_offset"],
                 self.parameters["x_tile"] * (x_index + 1) + self.parameters["x_offset"]]
        y_rng = [self.parameters["y_tile"] * y_index + self.parameters["y_offset"],
                 self.parameters["y_tile"] * (y_index + 1) + self.parameters["y_offset"]]
        z_rng = [z_index + self.parameters["z_offset"], z_index + 1 + self.parameters["z_offset"]]

        if z_index + self.parameters["z_offset"] < 0:
            data = np.zeros((self.parameters["x_tile"], self.parameters["y_tile"]),
                            dtype=np.int32, order="C")
        else:
            # Get data
            cnt = 0
            while cnt < 5:
                try:
                    data = self.remote.get_cutout(self.channel, self.parameters["resolution"],
                                                  x_rng, y_rng, z_rng, no_cache=True)
                    data = np.asarray(data, np.uint32)
                    break
                except Exception as err:
                    if cnt > 5:
                        raise err
                    cnt += 1
                    time.sleep(10)

        # Save sub-img to png and return handle
        upload_img = Image.fromarray(np.squeeze(data))
        output = six.BytesIO()
        upload_img.save(output, format="TIFF")

        # Send handle back
        return output



# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
import six
from PIL import Image
import numpy as np
from math import floor
import os
import re

from ..utils.filesystem import DynamicFilesystemAbsPath
from .path import PathProcessor
from .tile import TileProcessor


def load_tiff_multipage(tiff_filename, dtype='uint16'):
    """
    Load a multipage tiff into a single variable in x,y,z format.

    Arguments:
        tiff_filename:     Filename of source data
        dtype:             data type to use for the returned tensor

    Returns:
        Array containing contents from input tiff file in tyx order
    """
    if not os.path.isfile(tiff_filename):
        raise IOError('File not found: {}'.format(tiff_filename))

    # load the data from multi-layer TIF files
    data = Image.open(tiff_filename)

    im = []
    while True:
        # Get all slices from file
        img_slice = np.array(data, dtype=dtype)
        if img_slice.ndim == 2:
            img_slice = img_slice[np.newaxis, ...]
        im.append(img_slice)

        try:
            data.seek(data.tell()+1)
        except EOFError:
            break  # end of file

    im = np.concatenate(im, axis=0)
    return im


class SingleTimeTiffPathProcessor(PathProcessor):
    def setup(self, parameters):
        """Set the params - for this just store where the file is located

        MUST HAVE THE CUSTOM PARAMETERS: "z_<index>": "<filename>"

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        self.parameters = parameters

    def process(self, x_index, y_index, z_index, t_index=None):
        """
        Method to compute the file path for the indicated tile - For this, it's always the same file for each Z slice

        Args:
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (str): An absolute file path that contains the specified data

        """
        if t_index < self.parameters["ingest_job"]["extent"]["t"][0] or t_index >= self.parameters["ingest_job"]["extent"]["t"][1]:
            raise IndexError("Invalid Tile T-Index: {}".format(t_index))

        if z_index < self.parameters["ingest_job"]["extent"]["z"][0] or z_index >= self.parameters["ingest_job"]["extent"]["z"][1]:
            raise IndexError("Invalid Tile Z-Index: {}".format(z_index))

        if x_index > self.parameters["ingest_job"]["extent"]["x"][1] / self.parameters["ingest_job"]["tile_size"]["x"] - 1:
            raise IndexError("Invalid Tile X-Index: {}".format(x_index))

        if y_index > self.parameters["ingest_job"]["extent"]["y"][1] / self.parameters["ingest_job"]["tile_size"]["y"] - 1:
            raise IndexError("Invalid Tile Y-Index: {}".format(y_index))

        return self.parameters['z_{}'.format(z_index)]


class SingleTimeTiffTileProcessor(TileProcessor):
    """A Tile processor for a file where a multi-page TIFF contains all time points for a single z-slice"""
    def __init__(self):
        """Constructor to add custom class var"""
        TileProcessor.__init__(self)
        self.data = None

    def setup(self, parameters):
        """ Method to load the file for uploading - a very naive approach

        MUST HAVE THE CUSTOM PARAMETER: "datatype": "<uint8|uint16>"

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        self.parameters = parameters
        self.data = {}

    def process(self, file_path, x_index, y_index, z_index, t_index=0):
        """
        Method to load the configuration file and select the correct validator and backend

        Args:
            file_path(str): An absolute file path for the specified tile
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (io.BufferedReader): A file handle for the specified tile

        """
        # Load the file into memory
        if "z_{}".format(z_index) not in self.data:
            # storing slices in tyx
            self.data["z_{}".format(z_index)] = load_tiff_multipage(file_path, dtype=self.parameters["datatype"])

        im = self.data["z_{}".format(z_index)][t_index, :, :]

        # Compute matrix indices
        x_start = self.parameters["ingest_job"]["tile_size"]["x"] * x_index
        x_stop = self.parameters["ingest_job"]["tile_size"]["x"] * (x_index + 1)
        y_start = self.parameters["ingest_job"]["tile_size"]["y"] * y_index
        y_stop = self.parameters["ingest_job"]["tile_size"]["y"] * (y_index + 1)

        # TODO: Verify handles will be closed properly and memory reclaimed
        # Save img to png and return handle
        tile_data = Image.fromarray(im[y_start:y_stop, x_start:x_stop], 'I;16')

        output = six.BytesIO()
        tile_data.save(output, format="TIFF")

        # Send handle back
        return output


class TiffMultiFileHyperStackPathProcessor(PathProcessor):
    """A Path processor for a hyperstack stored across multiple multi-page TIFF files, with the time dimension split
    across files"""

    def __init__(self):
        """Constructor to add custom class var"""
        PathProcessor.__init__(self)
        self.regex = None

    def setup(self, parameters):
        """Set the params

        MUST HAVE THE CUSTOM PARAMETERS: "root_dir": "<path_to_stack_root>",
                                         "extension": "tiff|tif",
                                         "base_filename": the base filename, see below for how this is parsed,
                                         "time_chunk_size": <int>  # The number of time samples in a single file

        base_filename string identifies how to insert the z-index value into the filename. Identify a place to insert
        the z_index with "<>".  If you want to offset add o:number. If you want to zero pad add p:number"

        my_base_<> -> my_base_0, my_base_1, my_base_2
        <o:200>_my_base_<p:4> -> 200_my_base_0000, 201_my_base_0001, 202_my_base_0002

        Includes the "ingest_job" section of the config file automatically

        Args:
            parameters (dict): Parameters for the dataset to be processed

        Returns:
            None
        """
        self.parameters = parameters
        self.regex = re.compile('<(o:\d+)?(p:\d+)?>')

    def process(self, x_index, y_index, z_index, t_index=None):
        """
        Method to compute the file path for the indicated time sample

        Args:
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (str): An absolute file path that contains the specified data

        """
        # Create base filename
        matches = self.regex.findall(self.parameters['base_filename'])

        # Compute file number
        file_number = int(floor(t_index / int(self.parameters["time_chunk_size"])))

        base_str = self.parameters['base_filename']
        for m in matches:
            if m[0]:
                # there is an offset
                t_val = int(m[0].split(':')[1]) + file_number
            else:
                t_val = file_number

            if m[1]:
                # There is zero padding
                t_str = str(t_val).zfill(int(m[1].split(':')[1]))
            else:
                t_str = str(t_val)

            base_str = base_str.replace("<{}{}>".format(m[0], m[1]), t_str)

        # prepend root, append extension
        return os.path.join(self.parameters['root_dir'], "{}.{}".format(base_str, self.parameters['extension']))


class TiffMultiFileHyperStackTileProcessor(TileProcessor):
    """A Tile processor for multi-channel, multi-slice, time-series datasets stored as a hyperstack in a multi-page tiff
    """

    def __init__(self):
        """Constructor to add custom class var"""
        TileProcessor.__init__(self)
        self.fs = None

    def setup(self, parameters):
        """ Method to load the file for uploading

        Args:
            parameters (dict): Parameters for the dataset to be processed


        MUST HAVE THE CUSTOM PARAMETERS: "time_chunk_size": <int>,
                                         "num_z_slices": <int>,
                                         "num_channels": <int>,
                                         "channel_index": <int>,
                                         "filesystem": "<s3|local>",
                                         "bucket": (if s3 filesystem)

        Returns:
            None
        """
        self.parameters = parameters
        self.fs = DynamicFilesystemAbsPath(parameters['filesystem'], parameters)

    def process(self, file_path, x_index, y_index, z_index, t_index=0):
        """
        Method to load the image file.

        Args:
            file_path(str): An absolute file path for the specified tile
            x_index(int): The tile index in the X dimension
            y_index(int): The tile index in the Y dimension
            z_index(int): The tile index in the Z dimension
            t_index(int): The time index

        Returns:
            (io.BufferedReader): A file handle for the specified tile

        """
        file_path = self.fs.get_file(file_path)

        # Open Tiff Hyper-Stack
        tiff_file = Image.open(file_path)

        # Compute frame Number
        frame_num = ((self.parameters["num_z_slices"] * self.parameters["num_channels"]) * t_index) + \
                    (z_index * self.parameters["num_channels"]) + (self.parameters["channel_index"])

        tiff_file.seek(frame_num % self.parameters["time_chunk_size"])

        tile_data = np.array(tiff_file, dtype=np.uint16)
        upload_img = Image.fromarray(tile_data, 'I;16')

        output = six.BytesIO()
        upload_img.save(output, format="TIFF")

        # Send handle back
        return output
